import os
import subprocess
import glob
import pickle 
import shutil 


import pandas as pnd



def get_metadata_table(logger, rawmeta_filepath):
    
    
    # read the raw metadata: 
    logger.info("Creating the metadata table for your genomes...") 
    metadata = pnd.read_csv(rawmeta_filepath, index_col=0)
    
    
    # this table has 2 rows for each genome, one if the '_assembly_stats.txt' row.
    # here we delete the *assembly_stats rows: 
    to_drop = []
    for index, row in metadata.iterrows(): 
        if row.local_filename.endswith('_assembly_stats.txt'):
            to_drop.append(index)
    metadata = metadata.drop(to_drop)
    metadata = metadata.reset_index(drop=True)
    logger.debug("Shape of the metadata table: " + str(metadata.shape))
    
    
    # merge 'infraspecific_name' and 'isolate' to a single column 'strain_isolate': 
    metadata['infraspecific_name'] = metadata['infraspecific_name'].apply(lambda x: x.replace('strain=', '') if type(x)==str and x!='na' else '')
    metadata['isolate'] = metadata['isolate'].apply(lambda x: x if type(x)==str and x!='na' else '')
    metadata['strain_isolate'] = metadata['infraspecific_name'] + metadata['isolate']
    metadata = metadata.drop(['infraspecific_name', 'isolate'], axis=1)
    
    
    # select desired columns:
    # Warning: the same columns are used in handle_manual_genomes()
    metadata = metadata[['assembly_accession', 'bioproject', 'biosample', 'excluded_from_refseq', 'refseq_category', 'relation_to_type_material', 'species_taxid', 'organism_name', 'strain_isolate', 'version_status', 'seq_rel_date', 'submitter' ]] 
    
    
    # save the metadata table to disk:
    metadata.to_csv("working/genomes/genomes.csv")
    logger.info("Metadata table saved in ./working/genomes/genomes.csv.") 
    


def create_genomes_dictionary(logger): 
    
    
    # read the metadata table
    metadata = pnd.read_csv("working/genomes/genomes.csv", index_col=0)
    
    
    # create species-to-genome dictionary:
    species_to_genome = {}
    groups = metadata.groupby('organism_name').groups
    for species in groups.keys():
        indexes = groups[species]
        subset_metadata = metadata.iloc[indexes, ]
        species_to_genome[species] = [f'working/genomes/{accession}.fna' for accession in subset_metadata['assembly_accession']]
    logger.debug(f"Created the species-to-genome dictionary: {str(species_to_genome)}.") 
    
    
    # save the dictionary to disk: 
    with open('working/genomes/species_to_genome.pickle', 'wb') as file:
        pickle.dump(species_to_genome, file)
    logger.debug(f"Saved the species-to-genome dictionary to file: ./working/genomes/species_to_genome.pickle.")
    


def get_genomes(logger, taxids, cores): 
    
    
    # get metadata basename: 
    taxids_sorted = sorted(taxids.split(','))
    meta_basename = f"raw_ncbi_{'_'.join(taxids_sorted)}"
    
    
    # execute the download
    logger.info("Downloading from NCBI all the genome assemblies linked to the provided taxids...")
    with open('working/logs/stdout_download.txt', 'w') as stdout, open('working/logs/stderr_download.txt', 'w') as stderr: 
        command = f"""ncbi-genome-download \
            --no-cache \
            --metadata-table working/genomes/{meta_basename}.txt \
            --retries 100 --parallel 10 \
            --output-folder working/genomes/ \
            --species-taxids {taxids} \
            --formats assembly-stats,fasta \
            --section genbank \
            bacteria"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
    logger.debug("Download finished. Logs are stored in ./working/logs/stdout_download.txt and ./working/logs/stderr_download.txt.") 
    
    
    # format the metadata
    metadata = pnd.read_csv(f"working/genomes/{meta_basename}.txt", sep='\t')
    metadata.to_csv(f"working/genomes/{meta_basename}.csv")
    os.remove(f"working/genomes/{meta_basename}.txt")
    
    
    # moving the genomes to the right directory
    for file in glob.glob('working/genomes/genbank/bacteria/*/*.fna.gz'):
        accession = file.split('/')[-2]
        shutil.copy(file, f'working/genomes/{accession}.fna.gz')
    shutil.rmtree('working/genomes/genbank/') # delete the old tree
    logger.debug("Moved the downloaded genomes to ./working/genomes/.") 
    
    
    # execute the decompression
    logger.info("Decompressing the genomes...")
    with open('working/logs/stdout_decompression.txt', 'w') as stdout, open('working/logs/stderr_decompression.txt', 'w') as stderr: 
        command = f"""unpigz -p {cores} working/genomes/*.fna.gz""" 
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
    logger.debug("Decompression finished. Logs are stored in ./working/logs/stdout_decompression.txt and ./working/logs/stderr_decompression.txt.") 
    
    

def download_genomes(logger, taxids, cores):
    
    
    # create a sub-directory without overwriting
    os.makedirs('working/genomes/', exist_ok=True)
    
    
    # get metadata basename: 
    taxids_sorted = sorted(taxids.split(','))
    meta_basename = f"raw_ncbi_{'_'.join(taxids_sorted)}"
    
    
    # check if the genomes were already downloaded:
    if os.path.exists(f'working/genomes/{meta_basename}.csv'):
        metadata = pnd.read_csv(f"working/genomes/{meta_basename}.csv", index_col=0) 
        if all([os.path.exists(f'working/genomes/{accession}.fna') for accession in metadata['assembly_accession']]):
            metadata = metadata[metadata['local_filename'].str.endswith('assembly_stats.txt')==False]  # keep only rows for genomes
            logger.info(f"Genome assemblies already downloaded for taxids {taxids}: {len(metadata['assembly_accession'])} assemblies found. Skipping the download from NCBI.")
 
            # create metadata table and genomes dictionary: 
            get_metadata_table(logger, f'working/genomes/{meta_basename}.csv')
            create_genomes_dictionary(logger)


            return 0    
    
          
    # download from ncbi: 
    get_genomes(logger, taxids, cores)

    
    # create the metadata table and the genomes dictionary
    get_metadata_table(logger, f'working/genomes/{meta_basename}.csv')
    create_genomes_dictionary(logger)
    
    
    return 0 
    
    

def handle_manual_genomes(logger, genomes):
    
    
    # create a sub-directory without overwriting
    os.makedirs('working/genomes/', exist_ok=True)
    
    
    # create a species-to-genome dictionary
    species_to_genome = {}
    logger.debug(f"Checking the formatting of the provided -g/-genomes attribute...") 
    
    
    # check if the user specified a folder:
    if os.path.exists(genomes) and os.path.isdir(genomes):
            if genomes[-1] != '/': genomes = genomes + '/'
            files = glob.glob(genomes + '*')
            species_to_genome['Spp'] = files
    
    elif '+' in genomes and '@' in genomes: 
        for species_block in genomes.split('+'):
            species, files = species_block.split('@')
            for file in files.split(','): 
                if not os.path.exists(file):
                    logger.error("The following file provided in -g/--genomes does not exists: " + file)
                    return 1
            species_to_genome[species] = files.split(',')
            
    else: # the user has just 1 species
        for file in genomes.split(','): 
            if not os.path.exists(file):
                logger.error("The following file provided in -g/--genomes does not exists: " + file)
                return 1
        species_to_genome['Spp'] = genomes.split(',')

    
    # report a summary of the parsing: 
    logger.info(f"Inputted {len(species_to_genome.keys())} species with well-formatted paths to genomes.") 
    
    
    # move the genomes to the usual directory: 
    for species in species_to_genome.keys():
        copied_files = []
        for file in species_to_genome[species]:
            basename = os.path.basename(file)
            shutil.copyfile(file, 'working/genomes/' + basename)  # just the content, not the permissions. 
            copied_files.append('working/genomes/' + basename)
        species_to_genome[species] = copied_files
    logger.debug(f"Input genomes copied to ./working/genomes/.")
    logger.debug(f"Created the species-to-genome dictionary: {str(species_to_genome)}.") 
    
    
    # save the dictionary to disk: 
    with open('working/genomes/species_to_genome.pickle', 'wb') as file:
        pickle.dump(species_to_genome, file)
    logger.debug(f"Saved the species-to-genome dictionary to file: ./working/genomes/species_to_genome.pickle.")
    
    
    # Create the genomes/genomes.csv like if genomes were downloaded from NCBI.
    # Useful during plot generation.
    # Warning: the same columns are used in get_metadata_table(). But here only 2 can be filled: 'organism_name' and 'strain_isolate'.
    metadata = []  # list of dicsts, future df.
    for species in species_to_genome.keys(): 
        for genome in species_to_genome[species]:
            basename = os.path.basename(genome)
            accession, _ = os.path.splitext(basename)
            metadata.append({'assembly_accession': accession, 'strain_isolate': accession, 'organism_name': species})
    metadata = pnd.DataFrame.from_records(metadata)
    columns = ['assembly_accession', 'bioproject', 'biosample', 'excluded_from_refseq', 'refseq_category', 'relation_to_type_material', 'species_taxid', 'organism_name', 'strain_isolate', 'version_status', 'seq_rel_date', 'submitter' ]
    for col in columns:  # add missing columns: 
        if col not in metadata.columns: 
            metadata[col] = None
    # save the metadata table to disk:
    metadata.to_csv("working/genomes/genomes.csv")
    
    
    return 0
    


import re
import importlib.metadata


import cobra


from gempipe.curate.gaps import get_solver
from gempipe.curate.gaps import get_objectives


def close_boundaries(model):
    """Set all the EX_change reactions to (0, 0).
    
    Args:
        model (cobra.Model): target model.

    """
    
    for r in model.reactions:
        if len(r.metabolites)==1 and r.id.startswith('EX_'):
            r.bounds = (0, 0)



def verify_egc(model, mid, escher=False, threshold=1e-5, verbose=True): 
    """Test the presence of energy-generating cycles (EGCs). 
    
    Can also output a model for Escher, with just the reactions composing the cycle. 
    
    Args:
        model (cobra.Model): target model. Must be encoded with the BiGG notation.
        mid (str): metabolite ID for which the EGC must be checked. Warning: must be without compartment, so for example ``atp`` instead of ``atp_c``. 
        escher (bool): save a reduced ``cobra.Model`` in the current directory. To be loaded in Escher. 
        threshold (float): values below this treshold are considered as 0.
        verbose (bool): if True, show results of the test even if no EGC was detected.
        
    Returns:
        bool: ``True`` if an EGC is detected.
    """
    
    # changes as not permament: 
    with model: 
        
        # close all the exchange reactions: 
        close_boundaries(model)

                
        # create a dissipation reaction: 
        dissip = cobra.Reaction(f'__dissip__{mid}')
        model.add_reactions([dissip])
        dissip = model.reactions.get_by_id(f'__dissip__{mid}')
        
        
        # define the dissipation reaction:
        modeled_mids = [m.id for m in model.metabolites]
        if mid == 'atp':
            dissip_string = 'atp_c + h2o_c --> adp_c + pi_c + h_c'
        elif mid == 'ctp':
            dissip_string = 'ctp_c + h2o_c --> cdp_c + pi_c + h_c'
        elif mid == 'gtp':
            dissip_string = 'gtp_c + h2o_c --> gdp_c + pi_c + h_c'
        elif mid == 'utp':
            dissip_string = 'utp_c + h2o_c --> udp_c + pi_c + h_c'
        elif mid == 'itp':
            dissip_string = 'itp_c + h2o_c --> idp_c + pi_c + h_c'
        elif mid == 'nadh':
            dissip_string = 'nadh_c --> nad_c + h_c'
        elif mid == 'nadph':
            dissip_string = 'nadph_c --> nadp_c + h_c'
        elif mid == 'fadh2':
            dissip_string = 'fadh2_c --> fad_c + 2.0 h_c'
        elif mid == 'accoa':
            dissip_string = 'accoa_c + h2o_c --> ac_c + coa_c + h_c'
        elif mid == 'glu__L':
            if 'nh4_c' in modeled_mids :
                dissip_string = 'glu__L_c + h2o_c --> akg_c + nh4_c + 2.0 h_c'
            elif 'nh3_c' in modeled_mids :
                dissip_string = 'glu__L_c + h2o_c --> akg_c + nh3_c + 3.0 h_c'
            else:
                raise Exception("'nh4_c' or 'nh3_c' must be present in the model.")
        elif mid == 'q8h2':
            dissip_string = 'q8h2_c --> q8_c + 2.0 h_c'

        else: 
            raise Exception("Metabolite ID (mid) not recognized.") 
        dissip.build_reaction_from_string(dissip_string)
        
        
        # set the objective and optimize: 
        model.objective = f'__dissip__{mid}'
        res = model.optimize()
        
        
        # apply the threshold:
        obj_value = res.objective_value
        if abs(obj_value) < threshold:
            obj_value = 0
        
        
        # log some messages
        if obj_value != 0 or verbose:
            print(dissip.reaction)
            print(obj_value , ':', res.status )


            # get suspect !=0 fluxes (if any)
            if obj_value != 0: 
                fluxes = res.fluxes
                print()  # skip a line befor printing the EGC members

                # get interesting fluxes (0.001 tries to take into account the approximation in glpk and cplex solvers)
                fluxes_interesting = fluxes[((fluxes > 0.001) | (fluxes < -0.001)) & (fluxes.index != f'__dissip__{mid}')]
                print(fluxes_interesting.to_string())


                # create a model for escher
                if escher:  
                    model_copy = model.copy()
                    all_rids = [r.id for r in model_copy.reactions]
                    to_delete = set(all_rids) - set(fluxes_interesting.index)
                    model_copy.remove_reactions(to_delete)
                    cobra.io.save_json_model(model_copy, f'__dissip__{mid}' + '.json')
                    print(f'__dissip__{mid}' + '.json', "saved in current directory.")
                    
        if obj_value != 0: 
            return True
        else: return False



def verify_egc_all(model):
    """Quickly check the presence of EGCs for the main metabolites.
    
    Internally calls `verify_egc()` over a list of metabolite IDs.
    
    Args:
        model (cobra.Model): target model.
    """
    
    mids_to_check = ['atp','ctp','gtp','utp','itp','nadh','nadph','fadh2','accoa','glu__L','q8h2']
    all_results = []
    for mid in mids_to_check:
        result = verify_egc(model, mid, verbose=False)
        all_results.append(~result)
    if all(all_results):
        print("No energy-generating cycles (EGCs) found.")
        
        
                

    
def check_sink_demand(model, verbose=True):
    """Check presence of sink and demand reactions.
    
    Here they are simply defined as reactions involving just 1 metabolite, with an ID not starting with ``EX_``.
    
    Args:
        model (cobra.Model): target model.
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of sink/demand reactions found.
    """
    
    found_rids = []
    cnt = 0
    for r in model.reactions:
        if len(r.metabolites) == 1: 
            if r.id.startswith("EX_") == False: 
                cnt += 1
                if verbose: print(cnt, ':', r.id, ':', r.reaction, ':', r.bounds)
                found_rids.append(r.id)
    
    
    if found_rids == []:
        if verbose: print("No sink/demand reactions found.")
    return found_rids
   

                
def check_exr_notation(model, verbose=True): 
    """Check that every EX_change reaction ID begins with ``EX_``.
    
    Here EX_change reactions are defined as those reactions having just 1 metabolite involved, included in the extracellular compartment.
    
    Args:
        model (cobra.Model): target model. 
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of EX_change reactions with bad ID.
    """

    found_rids = []
    cnt = 0
    for r in model.reactions:
        if len(r.metabolites) == 1: 
            if list(r.metabolites)[0].id.rsplit('_', 1)[1] == 'e':  # extracellular compartment
                if r.id.startswith("EX_") == False: 
                    cnt += 1
                    if verbose: print(cnt, ':', r.id, ':', r.reaction)
                    found_rids.append(r.id)
    
    
    if found_rids == []:
        if verbose: print("No EX_change reaction with bad ID found.")
    return found_rids

 
    
def remove_EX_annots(model): 
    """Remove all annotations from EX_change reactions.
    
    Args:
        model (cobra.Model): target model. 
    """
    
    cnt = 0
    for r in model.reactions: 
        if len(r.metabolites) == 1 and r.id.startswith("EX_"): 
            if r.annotation != {}: 
                r.annotation = {}

    
    
def check_missing_charges(model, verbose=True):
    """Check if all metabolites have a charge attribute.
    
    Args:
        model (cobra.Model): target model. 
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of matabolites missing the charge attribute.
    
    """ 
    
    found_mids = []
    cnt = 0
    for m in model.metabolites:  
        cnt += 1
        charge = m.charge
        if charge==None or type(charge)!=int:
            if verbose: print(cnt, ':', m.id)
            found_mids.append(m.id)
            

    if found_mids == []:
        if verbose: print("No metabolite with missing charge attribute found.")
    return found_mids



def check_missing_formulas(model, verbose=True):
    """Check if all metabolites have a formula attribute.
    
    Args:
        model (cobra.Model): target model. 
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of matabolites missing the formula attribute.
    
    """ 
    
    found_mids = []
    cnt = 0
    for m in model.metabolites:  
        formula = m.formula
        if formula==None or formula=='':
            cnt += 1
            if verbose: print(cnt, ':', m.id)
            found_mids.append(m.id)
            

    if found_mids == []:
        if verbose: print("No metabolite with missing formula attribute found.")
    return found_mids



def check_artificial_atoms(model, preview=None, verbose=True):
    """Check if artificial atoms like 'R' and 'X' are present.
    
    Args:
        model (cobra.Model): target model. 
        preview (int): maximum number of metabolite IDs to show for each artificial atom. If ``None``, they will be all displayed.
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of matabolites with artificial atoms.
    """ 
    
    found_mids = []
    atom_to_mids = {}
    for m in model.metabolites:  
        formula = m.formula
        if formula == None or formula=='':
            continue  # there are dedicated functions for this
        
        
        # Matches any uppercase letter (A-Z) followed by zero or more lowercase letters (a-z)
        atoms = set(re.findall(r'[A-Z][a-z]*', formula))
        base = set(['C', 'H', 'O', 'N', 'P', 'S'])
        metals = set(['Ag', 'Fe', 'Co', 'As', 'Ca', 'Cd', 'Cl', 'Cu', 'Hg', 'K', 'Mg', 'Mo', 'Na', 'Ni', 'Se', 'Zn', 'Mn'])
        safe_atoms = base.union(metals)
        strange = atoms - safe_atoms
        
        
        # popupale te 'found_mids' list
        if len(strange) > 0: 
            found_mids.append(m.id)
        
        
        # iterate each 'strange' atom to create a dict: 
        for atom in strange: 
            if atom not in atom_to_mids.keys(): 
                atom_to_mids[atom] = []
            atom_to_mids[atom].append(f'{m.id}[{m.formula}]')
            
            
    # show the create 'atom_to_mids' dict: 
    for i, atom in enumerate(atom_to_mids.keys()): 
        mids_to_show = ', '.join(atom_to_mids[atom])
        if preview != None:
            if len(atom_to_mids[atom]) > preview:
                mids_to_show = f"{', '.join(atom_to_mids[atom][:preview])}, ... ({len(atom_to_mids[atom])} in total)"
        if verbose: print(i+1, ':', atom, ':', mids_to_show) 

        
    if found_mids == []:
        if verbose: print("No metabolite with artificial atoms found.")
    return found_mids



def get_unconstrained_bounds(model): 
    """Get the minimum and maximum bounds used in the model.
    
    Usually they correspond to the "unconstrained" negative and positive flux constants.
    
    Args:
        model (cobra.Model): target model. 
    
    Returns:
        tuple: min and max bounds.
    """
    un_lb, un_ub = 0, 0
    for r in model.reactions:
        if r.lower_bound < un_lb: un_lb = r.lower_bound
        if r.upper_bound > un_ub: un_ub = r.upper_bound
    return (un_lb, un_ub)



def reset_unconstrained_bounds(model): 
    """Set the uncontrained flux constants to 1000.
    
    Args:
        model (cobra.Model): target model. 
     
    """
    un_lb, un_ub = get_unconstrained_bounds(model)
    for r in model.reactions: 
        if r.lower_bound == un_lb: r.lower_bound = -1000
        if r.upper_bound == un_ub: r.upper_bound = 1000



def check_constrained_metabolic(model, verbose=True): 
    """Check the presence of constrained metabolic reactions.
    
    Metabolic reactions are here defined as those having more then 1 involved metabolites.
    Constrained reactions are here defined as those having bounds other then (0, 1000) or (-1000, 1000).
    
    Biomass assembly reactions are ignored.
    
    Args:
        model (cobra.Model): target model. 
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of reactions with constrained metabolic reactions.
    
    """
    
    found_rids = []
    cnt = 0
    biom_ids = search_biomass(model, verbose=False)
    for r in model.reactions: 
        if len(r.metabolites) == 1 or r.id in biom_ids: 
            continue # not interested in EXR/sink/demands, nor in biomass
            
        if r.bounds != (-1000, 1000) and r.bounds != (0, 1000): 
            cnt += 1
            if verbose: print(cnt, ':', r.id, ':', r.bounds, ':', r.reaction)
            found_rids.append(r.id)
    
    
    if found_rids == []:
        if verbose: print("No constrained metabolic reactions found.")
    return found_rids



def check_mass_unbalances(model, threshold=1e-5, verbose=True):
    """Check the presence of mass-unbalaned reactions in the model.
    
    EExchange, sink, demand, and biomass reactions are ignored.
    
    Args:
        model (cobra.Model): target model. 
        threshold (float): values below this treshold are considered as 0.
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of reactions with mass unbalanced.
    
    """
    
    found_rids = []
    cnt = 0
    biom_ids = search_biomass(model, verbose=False)
    for r in model.reactions:
        
        
        # exchage, sink, demand, and biomass reactions are excluded. 
        if len(r.metabolites)==1 or r.id in biom_ids: 
            continue 
        
        
        # get suggestions
        suggestions = r.check_mass_balance()
        
        # apply threshold: 
        suggestions = {atom: suggestions[atom] for atom in suggestions.keys() if suggestions[atom] >= threshold}
        
        # exclude charge is present:
        suggestions = {atom: suggestions[atom] for atom in list(set(list(suggestions.keys())) - set(['charge']))} 
        
        
        # log the suggestion:
        if len(suggestions.keys()) > 0:
            cnt += 1
            if verbose: print(cnt, ':', r.id, ':', r.reaction, ':', suggestions)
            found_rids.append(r.id)
    
    
    if found_rids == []:
        if verbose: print("No mass-unbalanced reactions found.")
    return found_rids



def check_charge_unbalances(model, threshold=1e-5, verbose=True):
    """Check the presence of charge-unbalaned reactions in the model.
    
    Exchange, sink, demand, and biomass reactions are ignored.
    
    Args:
        model (cobra.Model): target model. 
        threshold (float): values below this treshold are considered as 0.
        verbose (bool): if ``False``, don't print.
        
    Returns:
        list: IDs of reactions with charge unbalanced.
        
    """
    
    found_rids = []
    cnt = 0
    biom_ids = search_biomass(model, verbose=False)
    for r in model.reactions:
        
        
        # exchage, sink, demand, and biomass reactions are excluded. 
        if len(r.metabolites)==1 or r.id in biom_ids: 
            continue 
        
        
        # get suggestions
        suggestions = r.check_mass_balance()
        
        # retain only charge if present:
        suggestions = {atom: suggestions[atom] for atom in suggestions.keys() if atom == 'charge'} 
        
        # apply threshold: 
        suggestions = {atom: suggestions[atom] for atom in suggestions.keys() if suggestions[atom] >= threshold}
        
        
        # log the suggestion:
        if len(suggestions.keys()) > 0:
            cnt += 1
            if verbose: print(cnt, ':', r.id, ':', r.reaction, ':', suggestions)
            found_rids.append(r.id)
        
        
    if found_rids == []:
        if verbose: print("No charge-unbalanced reactions found.")
    return found_rids



def search_biomass(model, show_reaction=False, verbose=True):
    """Search for biomass reactions.
    
    Simple function involving just an exact match of substrings: 'biomass', 'growth', 'bof'.
    
    Args:
        model (cobra.Model): target model. 
        show_reaction (bool): whether to show also the reaction string
    
    """
    
    # define key objects: 
    substrings = ['biomass', 'growth', 'bof']
    found_rids = []
    cnt = 0
    
    
    def print_reaction(cnt, show_reaction, verbose):
        if verbose:
            if show_reaction: 
                print(cnt, ':', r.id, ':', r.name, ':', r.reaction, ':', r.bounds)
            else: 
                print(cnt, ':', r.id, ':', r.name, ':', r.bounds)
    
    
    # first try with reaction IDs:
    for r in model.reactions:
        for substring in substrings: 
            if substring.lower() in r.id.lower():
                cnt += 1
                print_reaction(cnt, show_reaction, verbose)
                found_rids.append(r.id)
    if found_rids != []:           
        return found_rids
                
        
    # then try with reaction names: 
    for r in model.reactions: 
        for substring in substrings: 
            if substring.lower() in r.name.lower():
                cnt += 1
                print_reaction(cnt, show_reaction, verbose)
                found_rids.append(r.id)
    if found_rids != []:           
        return found_rids
    
    
    # no biomass reactions found: 
    if found_rids == []:
        if verbose: print("No biomass reactions found.")
    return found_rids
    
    

def sanity_report(model):
    """Print a small sanity report.
    
    Get a sanity report calling some of the gempipe.curate functions.
    
    Args:
        model (cobra.Model): target model. 
    
    """
    
    version = importlib.metadata.metadata("gempipe")["Version"]
    print(f"gempipe v{version} - sanity_report")
    
    print("model ID:", model.id)
    
    G = len(model.genes)
    R = len(model.reactions)
    M = len(model.metabolites)
    uM = len(set([m.id.rsplit('_', 1)[0] for m in model.metabolites]))
    groups = len(model.groups)
    print('G:', G, 'R:', R, 'M:', M, 'uM:', uM, 'groups:', groups)
    
    comps = [c for c in model.compartments]
    print("Compartments:", sorted(comps))
    
    biom_ids = search_biomass(model, verbose=False)
    print("Biomass assemblies:", len(biom_ids), biom_ids)
    
    obj_ids = get_objectives(model)
    print("Objectives:", len(obj_ids), obj_ids)
    
    res = model.optimize()
    print("Optimization:", res.objective_value, res.status, f'({get_solver(model)})')
    
    print("Unconstrained LB-UB:", get_unconstrained_bounds(model))
    
    print("Bad EX_change notation:", len(check_exr_notation(model, verbose=False)))
    
    print("Sink/demand:", len(check_sink_demand(model, verbose=False)))
    
    print("Constrained metabolic:", len(check_constrained_metabolic(model, verbose=False)))
    
    print("With 'artificial' atoms:", len(check_artificial_atoms(model, verbose=False)))
    
    print(
        "Missing formulas - charges:",
        len(check_missing_formulas(model, verbose=False)), '-',
        len(check_missing_charges(model, verbose=False)),
    )
    
    print(
        "Mass - charge unbalances:",
        len(check_mass_unbalances(model, verbose=False)), '-',
        len(check_charge_unbalances(model, verbose=False)),
    )
    


import os
from importlib import resources
import subprocess
import shutil


import pandas as pnd
import cobra


from .networkrec import get_gene_scores_table
from .networkrec import get_protein_scores_table
from .networkrec import get_reaction_scores_table
from .networkrec import normalize_reaction_scores
from .networkrec import get_universe_template



from ..commons import get_blast_header




def run_tcdb_aligner(logger, cores):
    
    
    # some log messages
    logger.debug("Copying the TCDB database...")
    with resources.path("gempipe.assets", "tcdb_proteins.dmnd") as asset_path:  
        shutil.copyfile(asset_path, 'working/tcdb_transporters/tcdb_proteins.dmnd')
    
    
    # some log messages
    logger.debug("Aligning to the TCDB genes database...")
    # run the command:
    with open(f'working/logs/stdout_tcdbalign.txt', 'w') as stdout, open(f'working/logs/stderr_tcdbalign.txt', 'w') as stderr: 
        command = f"""diamond blastp --threads {cores} \
            -d working/tcdb_transporters/tcdb_proteins.dmnd \
            -q working/annotation/representatives.faa \
            -o working/tcdb_transporters/tcdb_alignment.tsv \
            --ultra-sensitive --quiet \
            --outfmt 6 {get_blast_header()}"""
        # not "--top 10" to avoid competing families.
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()



def filter_alignment(logger):
    
    
    header = get_blast_header().split(' ')
    alignment = pnd.read_csv('working/tcdb_transporters/tcdb_alignment.tsv', sep='\t', names=header)
    
    identity = 45
    positivity = 60
    coverage = 80
    query = f'pident >= {identity} & ppos >= {positivity} & qcovhsp >= {coverage} & scovhsp >= {coverage}'
    alignment_filtered = alignment.query(query)
    
    logger.debug(f"HSPs reduction: {len(alignment)} -> {len(alignment_filtered)}.")
    
    return alignment_filtered



def tcdb_matching(logger, tcdb_rs, reaction_scores_normalized):
    
    
    # get transport reactions completely supported by genes:
    rids_normalized = set(list(reaction_scores_normalized['reaction']))
    logger.debug(f"Number of retrieved transporters: {len(rids_normalized)}")


    # get transport reactions with chebi annotation:
    rids_chebi = set(list(tcdb_rs[tcdb_rs['rid'].isin(rids_normalized)]['rid']))
    logger.debug(f'With a "chebi substrate annotation: {len(rids_chebi)}/{len(rids_normalized)}')
    logger.debug(f"Missing annotation for {len(rids_normalized - rids_chebi)}: {[rid[2:].replace('_','.') for rid in (rids_normalized - rids_chebi)]}")


    # compute intersection: 
    intersection = rids_chebi.intersection(rids_normalized)
    matched = tcdb_rs[tcdb_rs['rid'].isin(intersection)]
    matched = matched.reset_index(drop=True)
    matched['gpr'] = None
    for index, row in matched.iterrows(): 
        matched.loc[index, 'gpr'] = reaction_scores_normalized[reaction_scores_normalized['reaction']==row['rid']]['GPR'].iloc[0]


    return matched



def add_putative_transporters(matched, model, universe):
    
    
    matched['added'] = None
    for index, row in matched.iterrows(): 
        
        
        chebi_2_added_rstrings = {}
        
        # get mids of involved metabolites: 
        for chebi_id, rstrings in eval(row['rstrings']).items():
            
            
            # each Chebi-ID is linked to putative transport reactions (rstrings)
            # (can be more then 1 reactions, as a Chebi-ID could be associated with more then 1 BiGG-IDs).
            # With 'added_rstrings' we keep trace of which rstrings have been added to the model for each Chebi-ID. 
            # 'chebi_2_added_rstrings' will be added to each row. 
            added_rstrings = []
            
            
            for i, rstring in enumerate(rstrings): 
                mids_involved = rstring.replace(' --> ', ' + ').replace(' <=> ', ' + ')
                mids_involved = mids_involved.split(' + ')
                
                
                # check if all 'mids_involved', are already included in the model:
                missing = set()
                for mid in mids_involved: 
                    try: m = model.metabolites.get_by_id(mid)
                    except: missing.add(mid)
                    
                    
                # if at least one involved metabolite cannot be recovered from panmodel/universe,
                # then interrupt the addition of this 'rstring' setting 'to_skip=True'
                to_skip = False
                
                    
                # check if missing mids are already modeled, but in another compartment
                for mid in missing: 
                    puremid = mid.rsplit('_', 1)[0]
                    found_m = None
                    for comp in ['c', 'p', 'e']:
                        try: found_m = model.metabolites.get_by_id(puremid + f'_{comp}')
                        except: pass
                    
                    
                    # if not found, check the universe
                    if found_m == None:
                        
                        for comp in ['c', 'p', 'e']:
                            try: found_m = universe.metabolites.get_by_id(puremid + f'_{comp}')
                            except: pass
                        
                        
                        # if missing from the universe too, interrupt the addition of this 'rstring': 
                        if found_m == None:
                            to_skip = True
                            break
                            
                            
                    # the misssing metabolite was found:
                    m = cobra.Metabolite(mid)
                    m.formula = found_m.formula
                    m.charge = found_m.charge
                    m.compartment = mid.rsplit('_', 1)[-1]
                    
                    
                    # add the missing metabolite: 
                    model.add_metabolites([m])
                          
                        
                # if at least one involved metabolite cannot be recovered from panmodel/universe,
                # then interrupt the addition of this 'rstring' setting 'to_skip=True'
                if to_skip:
                    added_rstrings.append(False)
                    break
                    
                    
                # at this point, all involved metabolites are present in the model. 
                # So we can add the reaction.
                # 'rid' will be TCDB-ID ('.' to '_') + Chebi-ID (as many transporters have specificity for more then one substrate)
                rid = row['rid'] + '__' + str(chebi_id) + '__' + str(i)
                r = cobra.Reaction(rid)
                model.add_reactions([r])
                r = model.reactions.get_by_id(rid)
                
                
                # manage rstring
                r.build_reaction_from_string(rstring)
                
                
                # manage gpr
                r.gene_reaction_rule = row['gpr']
                r.update_genes_from_gpr()
                
                
                # manage bounds
                if row['rev']=='-->' : r.bounds = (0, 1000) 
                else: r.bounds = (-1000, 1000) 
                
                
                added_rstrings.append(True)
            chebi_2_added_rstrings[chebi_id] = added_rstrings
        matched.loc[index, 'added'] = str(chebi_2_added_rstrings)                                 
                                       
        
    return model, matched



def stats_tcdb(logger, matched, plot=False): 
    
    
    # number of transporters
    n_ts = len(matched)
    
    
    # number of transporters annotated with chebi 
    n_annotated = len([eval(i) for i in matched['chebi'].to_list()  if eval(i) != {}])
    
    
    # number of transporters with at least 1 bigg reaction
    n_rstrings = 0
    for index, row in matched.iterrows(): 
        increment = False
        for chebi_id, rstrings in eval(row['rstrings']).items():
            if rstrings != set():
                increment = True
        if increment: 
            n_rstrings += 1
                
                
    # number of transporters with at least 1 bigg reaction INSERTED
    n_inserted = 0
    for index, row in matched.iterrows(): 
        increment = False
        for chebi_id, added in eval(row['added']).items():
            for response in added: 
                if response == True:
                    increment = True
        if increment: 
            n_inserted += 1
            
            
    # print statistics: 
    logger.info(f"n_ts: {n_ts} | n_annotated: {n_annotated} | n_rstrings: {n_rstrings} | n_inserted: {n_inserted}")
    
    
    # make the barplot:
    if plot: 
        categories = ['n_ts', 'n_annotated', 'n_rstrings', 'n_inserted']
        values = [n_ts, n_annotated, n_rstrings, n_inserted]  
        colors = ['C0', 'C1', 'C2', 'C3']
        fig, ax = plt.subplots()
        bars = ax.bar(categories, values, color=colors)
        [ax.text(bar.get_x() + bar.get_width() / 2, bar.get_height(), f'{bar.get_height()}', ha='center', va='bottom') for bar in bars]
        ax.set_xlabel('')
        ax.set_ylabel('')
        ax.set_title(f'modeled tranporters: {round(n_inserted/n_ts*100, 1)}%')


        return fig
    
    
    
def add_missing_exchanges(logger, model):
    
    added_exr_rids = []
    for m in model.metabolites: 
        if m.id.endswith('_e'):
            try: r = model.reactions.get_by_id(f'EX_{m.id}')
            except: 
    
                # create the reaction
                adding = cobra.Reaction(f"EX_{m.id}")
                adding.name = f"Exchange for {m.name}"
            
                # effectively add the reaction
                model.add_reactions([adding])
                added = model.reactions.get_by_id(f"EX_{m.id}")

                # define the manually corrected reaction:
                r_string = f"{m.id} -->"
                added.build_reaction_from_string(r_string) 

                # set bounds (just for EX_reactions we specify bounds):
                added.bounds = (0, 1000)
                
                added_exr_rids.append(f"EX_{m.id}")
                
    # some log messages:
    if logger != None: logger.debug(f"Added {len(added_exr_rids)} new EXR: {added_exr_rids}")
    
    return added_exr_rids



def tcdbing_main(logger, cores, staining):
    
    
    # create subdirs without overwriting
    os.makedirs('working/tcdb_transporters/', exist_ok=True)
    
    
    # some log messages
    logger.info('Experimental feature requested: trying to build transport reactions using TCDB...')
    
    
    # check if the needed files are already computed: 
    if False:
        logger.info('Found all the needed files already computed. Skipping this step.')
        # signal to skip this module:
        return 0
        
    
    # align representatives sequences on TCDB genes:
    run_tcdb_aligner(logger, cores)
    
    
    # filter the alignment
    alignment_filtered = filter_alignment(logger)

    
    # get the 'gprm_table':
    with resources.path("gempipe.assets", "tcdb_gprs.csv") as asset_path:  
        gprm_table = pnd.read_csv(asset_path)
    gprm_table.to_csv('working/tcdb_transporters/gprm_table.csv')
    
    
    # get the 'gene_scores' table:
    gene_scores = get_gene_scores_table(logger, alignment_filtered, gprm_table)
    gene_scores.to_csv('working/tcdb_transporters/gene_scores.csv')
    
    
    # get the 'protein_scores' table: 
    protein_scores = get_protein_scores_table(logger, gene_scores)
    protein_scores.to_csv('working/tcdb_transporters/protein_scores.csv')
    
    
    # get the 'reaction_scores' table: 
    reaction_scores = get_reaction_scores_table(logger, protein_scores)
    reaction_scores.to_csv('working/tcdb_transporters/reaction_scores.csv')
    
    
    # normalize reaction scores:
    reaction_scores_normalized = normalize_reaction_scores(reaction_scores)
    reaction_scores_normalized.to_csv('working/tcdb_transporters/reaction_scores_normalized.csv')
    
    
    # matching alingment with precomputed reaction database ('tcdb_rs'):
    with resources.path("gempipe.assets", "tcdb_rs.csv") as asset_path:  
        tcdb_rs = pnd.read_csv(asset_path)
    matched = tcdb_matching(logger, tcdb_rs, reaction_scores_normalized)
    matched.to_csv('working/tcdb_transporters/matched.csv', index=False)
    
    
    # using 'matched', expand the reference draft pan-model.
    # The resulting model will REPLACE the 'working/duplicates/draft_panmodel.json'.
    # load draft panmodel:
    panmodel = cobra.io.load_json_model('working/duplicates/draft_panmodel.json')
    uni = get_universe_template(logger, staining)
    logger.debug("Adding putative transporters...")
    panmodel, matched_added = add_putative_transporters(matched, panmodel, uni)
    matched_added.to_csv('working/tcdb_transporters/matched_added.csv')
    logger.debug("Adding missing exchanges...")
    add_missing_exchanges(logger, panmodel)
    logger.debug("Replacing working/duplicates/draft_panmodel.json...")
    cobra.io.save_json_model(panmodel, 'working/duplicates/draft_panmodel.json')
                                        
    
    # print some statistics:
    stats_tcdb(logger, matched_added)
    
    
    return 0


import os
import shutil 
import copy
import pickle


import pandas as pnd
import cobra


from ..commons import get_retained_accessions
from ..commons import get_md5_string



def check_same_mids(r, ref_r, remove_h=False): 
    
    
    # get metabolites and products of each reaction: 
    reacs = [m.id for m in r.reactants]
    prods = [m.id for m in r.products]
    ref_reacs = [m.id for m in ref_r.reactants]
    ref_prods = [m.id for m in ref_r.products]
    
    
    # if requested, do not consider protons:
    if remove_h: 
        reacs  = [mid for mid in reacs if mid.rsplit('_', 1)[0] != 'h']
        prods  = [mid for mid in prods if mid.rsplit('_', 1)[0] != 'h']
        ref_reacs  = [mid for mid in ref_reacs if mid.rsplit('_', 1)[0] != 'h']
        ref_prods  = [mid for mid in ref_prods if mid.rsplit('_', 1)[0] != 'h']
        

    # test both ways as reactions could be written upside down
    if   set(reacs)==set(ref_reacs) and set(prods)==set(ref_prods): 
        return True
    elif set(reacs)==set(ref_prods) and set(prods)==set(ref_reacs): 
        return True
    else: 
        return False
    
    
    
def check_same_fc(r, refmodel, remove_h=False):

    
    # parse each metabolite of this reaction
    comparison = []
    for m in r.metabolites: 
        
        
        # if requested, do not consider protons:
        if remove_h:
            if m.id.rsplit('_', 1)[0] == 'h':
                continue
        
        
        # check if same molecular formula and charge: 
        ref_m = refmodel.metabolites.get_by_id(m.id)
        if (m.formula == ref_m.formula) and (m.charge == ref_m.charge):
            comparison.append(True)
        else: 
            comparison.append(False)
    return all(comparison)



def check_same_gids(r, ref_r):

    
    # get the gene sets
    gids = set([g.id for g in r.genes])
    ref_gids = set([g.id for g in ref_r.genes])
    
    
    # check if two reactions have the same gene set in their gprs.
    if gids == ref_gids: 
        return True
    else: return False



def get_glued_gpr(r, ref_r):
    
    
    # get the gene sets
    gids = set([g.id for g in r.genes])
    ref_gids = set([g.id for g in ref_r.genes])
    
    
    # glue together the gprs from two different reactions: 
    if len(gids) == 0 and len(ref_gids) == 0:
        return ''
    elif len(gids) == 0: 
        return ref_r.gene_reaction_rule
    elif len(ref_gids) == 0:
        return r.gene_reaction_rule
    else: 
        return f'({r.gene_reaction_rule}) or ({ref_r.gene_reaction_rule})'

    
    
def search_first_synonym(r, refmodel, remove_h=False ): 
    
    
    # check if the reference contains an equal reaction based on reactants and products mids. 
    for ref_r in refmodel.reactions:
        found_synonym = check_same_mids(r, ref_r, remove_h)
        if found_synonym: 
            return ref_r

    
    # pay attention if ATPM / NGAM is returned !
    return None
    
    
    
def get_detailed_reaction(r): 
    
    
    # make esplicit formula and charge after each metabolite
    reaction = r.reaction
    reaction = ' ' + reaction + ' '  # force each mid to be surrounded by spaces
    for m in r.metabolites:
        reaction = reaction.replace(f' {m.id} ', f' {m.id}[{m.formula},{m.charge}] ')
    reaction = reaction[1:-1]  # remove extreme spaces
    
    
    # include balancing information
    if r.check_mass_balance() != {}:
        reaction = '[[UNB]] ' + reaction
    
    return reaction 
    


def add_new_reaction(r, refmodel, addedms_logger, mancor, empty_gpr=False, bounds=None):
    
    
    # don't edit the model if r.id appears in blacklist:
    if r.id in mancor['blacklist']:
        return 'blacklist'  # action for the dataframe
    
    
    # Notes for manual correction of fatty-acid-ACP formulas: 
    # Take as example hexACP: 
    # C 390 H 613 O 143 N 96 P 1 S 3    (-1) <--- proposed
    # C 6 H 11 O X     (0) <--- desired 
    # C 384 H 602 O 142 N 96 P 1 S 3     <--- to subtract
    
    
    # get metabolites already modeled
    reference_mids = set([m.id for m in refmodel.metabolites])
    
    
    # check if the reference model is purely based on stoichiometry:
    pure_stoich = True
    for m in refmodel.metabolites:
        if m.charge != None or m.formula != None: 
            pure_stoich = False
            break
    
    
    # first add missing metabolites:
    for m in r.metabolites: 
        if m.id not in reference_mids: 
            
            
            # create a new metabolite copying the attributes. 
            new_m = cobra.Metabolite(m.id)
            new_m.name = m.name
            
            # define formula and charge (here some manual corrections can be made)
            pure_mid = m.id.rsplit('_', 1)[0]
            new_m.formula = m.formula if pure_mid not in mancor['formulas'].keys() else mancor['formulas'][pure_mid]
            new_m.charge = m.charge if pure_mid not in mancor['charges'].keys() else mancor['charges'][pure_mid]
            
            # if the reference model is purely based on stoichoimetry, then we 
            # do not want to ruin its Memote metrics adding new metabolites with charge/formula:
            if pure_stoich: 
                new_m.formula = None
                new_m.charge = None
            
            # actually add the metabolite
            refmodel.add_metabolites([new_m])
            new_m = refmodel.metabolites.get_by_id(m.id)
            
            # set compartment:
            new_m.compartment = new_m.id.rsplit('_', 1)[-1]   
            
            print(f'id="{new_m.id}"\tformula="{new_m.formula}"\tcharge="{new_m.charge}"\tname="{new_m.name}"', file=addedms_logger)
            
    
    # now add missing reaction: 
    # first create a new empty reaction
    new_r = cobra.Reaction(r.id)
    new_r.name = r.name
    
    # actually add the reaction
    refmodel.add_reactions([new_r])
    new_r = refmodel.reactions.get_by_id(r.id)
    
    # define the reaction (here some manual curations can be made)
    new_r_reaction = r.reaction if r.id not in mancor['reactions'].keys() else mancor['reactions'][r.id]
    new_r.build_reaction_from_string(new_r_reaction) 
    
    # set bounds
    if bounds != None: new_r.bounds = bounds
    else: new_r.bounds = r.bounds
    
    # copy GPR: 
    if not empty_gpr:
        new_r.gene_reaction_rule = r.gene_reaction_rule 
        new_r.update_genes_from_gpr()
    
    
    return 'add' # action for the dataframe 
    
    

def mancor_to_dict(logger, mancor): 
    
    
    # convert the manual corrections file to dict: 
    resdict = {'formulas': {}, 'charges': {}, 'reactions': {}, 'blacklist': []}
    with open(mancor, 'r') as file: 
        for line in file.readlines(): 
            
            line = line.strip().rstrip()
            if line == '': continue
            if line.startswith('%'): continue  # commented line
            
            
            if line.startswith('formula.'):
                line = line[len('formula.'):]
                key, value = line.split(':', 1)
                resdict['formulas'][key] = value  
                
            elif line.startswith('charge.'):
                line = line[len('charge.'):]
                key, value = line.split(':', 1)
                value = int(value)
                resdict['charges'][key] = value
                
            elif line.startswith('reaction.'):
                line = line[len('reaction.'):]
                key, value = line.split(':', 1)
                resdict['reactions'][key] = value     
                
            elif line.startswith('blacklist.'):
                key = line[len('blacklist.'):]
                resdict['blacklist'].append(key)              
                
            else: 
                logger.error(f"Not recognized key in this line of the manual corrections file (-m/--mancor): {line}.")
                return 1

    
    return resdict

 
def expand_reference(refmodel, draft_panmodel, mancor): 
    
    
    # begin the addition of new reactions to the refmodel, to form the final draft pan-model.
    reference_rids = [r.id for r in refmodel.reactions]
    results_df = []  # summarizing all the additions
    addedms_logger = open('working/expansion/added_metabolites.txt', 'w')
    for r in draft_panmodel.reactions:
        if r.id == 'Growth': continue  # using reference biomass definition
        if [g.id for g in r.genes] == ['spontaneous']: continue  # not interested in spontaneous reactions.
        gpr = r.gene_reaction_rule
        reaction = get_detailed_reaction(r)
        
        
        # define key objects: 
        action = '-'
        rid_found = '-'
        same_mids = '-'
        same_fc = '-'
        same_gids = '-'
        ref_reaction = '-'
        final_reaction = '-'
        ref_gpr = '-'
        final_gpr = '-'
        synonym = '-'
        bal_suggestions = '-'
        
        
        # check if this rid is already modeled: 
        rid_found = r.id in reference_rids
        if rid_found:
            # get the reference reaction and gpr:
            ref_r = refmodel.reactions.get_by_id(r.id)
            
            
        else: # this rid is not modeled yet
            # search for synonyms in refmodel and take the first:
            ref_r = search_first_synonym(r, refmodel, remove_h=True)
            if ref_r != None: 
                synonym = ref_r.id
        
        
        # manage the reference reaction if any:
        if ref_r != None:
            ref_gpr = ref_r.gene_reaction_rule
            ref_reaction = get_detailed_reaction(ref_r)
            final_reaction = ref_reaction
            bal_suggestions = str(ref_r.check_mass_balance())
            
            
            # check if reactants and products are the same: 
            same_mids = check_same_mids(r, ref_r, remove_h=True)
            if same_mids: 
                # check if also formula and charge correspond: 
                same_fc = check_same_fc(r, refmodel, remove_h=True)
                
                
            # check if the gpr is the same, and define final gpr: 
            same_gids = check_same_gids(r, ref_r)
            if not same_gids: 
                # if different, glue together the two gprs, and update the reference:  
                action = 'update_gpr'
                final_gpr = get_glued_gpr(r, ref_r)
                ref_r.gene_reaction_rule = final_gpr  
                ref_r.update_genes_from_gpr()
            else: # if same gene set, just keep the definition from the reference 
                action = 'ignore'
                final_gpr = ref_gpr
                             
                
        else:  # new reaction to be included:
            
            action = add_new_reaction(r, refmodel, addedms_logger, mancor)
            if action == 'add':  # not in blacklist 
                new_r = refmodel.reactions.get_by_id(r.id)
                final_reaction = get_detailed_reaction(new_r)
                bal_suggestions = str(new_r.check_mass_balance())
                
                
                """
                from memote.support.consistency import find_unconserved_metabolites 
                cobra.io.save_json_model(refmodel, '/home/jovyan/work/validation/refmodel.json')
                refmodel2 = cobra.io.load_json_model('/home/jovyan/work/validation/refmodel.json')
                print(new_r.id, find_unconserved_metabolites(refmodel), find_unconserved_metabolites(refmodel2), len(refmodel.reactions), len(refmodel2.reactions))
                """
                
         
        
        # populate the results dataframe: 
        if bal_suggestions == '{}': bal_suggestions = '-'   # put back to standard formatting
        if r.id.startswith('EX_'): bal_suggestions = '-'   # non-sense for EX_ reactions
        new_row = {
            'rid': r.id, 'action': action, 
            'reaction': reaction, 'rid_found': rid_found, 'synonym': synonym, 'ref_reaction': ref_reaction,
            'same_mids': same_mids, 'same_fc': same_fc, 'final_reaction': final_reaction, 'bal_suggestions': bal_suggestions,
            'same_gids': same_gids, 'gpr': gpr, 'ref_gpr': ref_gpr, 'final_gpr': final_gpr, }
        results_df.append(new_row)
    
    
    # save results dataframe to disk
    addedms_logger.close()
    results_df = pnd.DataFrame.from_records(results_df)
    results_df.to_csv('working/expansion/results.csv')
    return results_df

    
    
def ref_expansion(logger, refmodel, mancor, identity, coverage): 
    
    
    # log some message
    logger.info("Expanding the reference model with new reactions taken from the reference-free reconstruction...")
    
    
    # create sub-directories without overwriting:
    os.makedirs('working/expansion/', exist_ok=True)
    
    
    # check the existence of the manual corrections file:
    mancor_filepath = copy.deepcopy(mancor)
    if mancor_filepath != '-': 
        if not os.path.exists(mancor_filepath): # check the input:
            logger.error(f"Provided path to the manual corrections (-m/--mancor) does not exist: {mancor_filepath}.")
            return 1
        
        
    # do the md5: 
    if mancor_filepath != '-': # existence of the file was verified above
        md5_new = get_md5_string(mancor_filepath)
    else: 
        md5_new = '-'
            
    
    # check if the output was already computed
    if os.path.exists('working/expansion/proc_acc.pickle'):
        with open('working/expansion/proc_acc.pickle', 'rb') as handler:
            proc_acc = pickle.load(handler) 
        if get_retained_accessions() == proc_acc:
            if os.path.exists(f'working/expansion/mancor.txt'):
                md5_old = get_md5_string('working/expansion/mancor.txt')
            else: md5_old = '-'  # fresh run , or previous run without mancor.
            if md5_old == md5_new:
                if os.path.exists(f'working/expansion/draft_panmodel.json'):
                    if os.path.exists(f'working/expansion/results.csv'):
                        if os.path.exists(f'working/expansion/added_metabolites.txt'):
                            logger.info('Found all the needed files already computed. Skipping this step.')
                            # signal to skip this module:
                            return 0
    
    
    # get the manual corrections dictionary:
    if mancor_filepath != '-': # existence of the file was verified above
        response = mancor_to_dict(logger, mancor_filepath)  # convert the filepath to dict
        if type(response) == int:
            if response == 1: return 1
        else: # good corrections file: 
            mancor = response  
            logger.info(  # log some message
                f"Using the provided manual corrections ({len(mancor['formulas'].keys())} formula corrections, {len(mancor['charges'].keys())} charge corrections, " + \
                f"{len(mancor['reactions'].keys())} reaction corrections, and {len(mancor['blacklist'])} reactions in blacklist)...")
    else: mancor = {'formulas': {}, 'charges': {}, 'reactions': {}, 'blacklist': []}  # emtpy, easier to handle

    
    # load the reference and the reference-free reconstruction.
    refmodel_basename = os.path.basename(refmodel)
    refmodel = cobra.io.load_json_model(f'working/brh/{refmodel_basename}.refmodel_translated.json')
    draft_panmodel = cobra.io.load_json_model(f'working/free/draft_panmodel_{identity}_{coverage}.json')
    
    
    # perform the main task
    results_df = expand_reference(refmodel, draft_panmodel, mancor)
    draft_panmodel_exp = refmodel  # after the expansion
    logger.info(f"Done, {' '.join(['G:', str(len(draft_panmodel_exp.genes)), '|', 'R:', str(len(draft_panmodel_exp.reactions)), '|', 'M:', str(len(draft_panmodel_exp.metabolites))])}.")
    
    
    # count how many unbalanced reactions remained
    unbalanced = results_df[results_df['rid'].str.startswith('EX_')==False]
    unbalanced = unbalanced[unbalanced['final_reaction'].str.startswith('[[UNB]] ')]
    logger.info(f"Terminated with {len(unbalanced)} unbalanced reactions (excluding EX_ reactions).")
    
    
    # finally save the new draft panmodel
    cobra.io.save_json_model(draft_panmodel_exp, 'working/expansion/draft_panmodel.json')
    logger.debug("New draft pan-model saved to 'working/expansion/draft_panmodel.json'.")
    
    
    # make traces to keep track of the accessions processed:
    # copy the one of brh/ , since the translated refmodelis the starting point
    shutil.copyfile('working/brh/proc_acc.pickle', 'working/expansion/proc_acc.pickle')
    if mancor_filepath != '-': shutil.copyfile(mancor_filepath, 'working/expansion/mancor.txt')
    else:  # delete an eventual previous version: 
        if os.path.exists(f'working/expansion/mancor.txt'):
            os.remove(f'working/expansion/mancor.txt')
    
    
    return 0



import multiprocessing
import itertools
import os
import shutil


import cobra


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results



def subtract_clusters(ss_model, panmodel, pam, accession):
    
    
    modeled_gids = [g.id for g in panmodel.genes]  # get medoled genes ID
    to_remove = []  # genes to delete
    
    
    # iterate the PAM :
    for cluster in pam.index: 
        # consider only if it is modeled:
        if cluster in modeled_gids: 
            cell = pam.loc[cluster, accession]
            if type(cell) == float:  # empty cell
                to_remove.append(ss_model.genes.get_by_id(cluster))
                continue
            # get all the sequences not containing a premature stop:
            seqs = [i for i in cell.split(';') if i.endswith('_stop')==False]
            if len(seqs) == 0:  # they were all '_stop' sequences.
                to_remove.append(ss_model.genes.get_by_id(cluster))
                continue
    
    
    # delete marked genes
    cobra.manipulation.delete.remove_genes(ss_model, to_remove, remove_reactions=True)
    
    
    return 0



def translate_genes(logger, ss_model, panmodel, pam, accession):
    
    
    # create a dups dict: 
    dups_dict = {}  # {'Cluster_XXXX': ['GFDSA_XXXX', 'GFDSA_XXXX', 'GFDSA_XXXX']}
    modeled_gids = [g.id for g in panmodel.genes]  # get medoled genes ID
    
    
    # iterate the PAM :
    for cluster in pam.index: 
        # consider only if it is modeled:
        if cluster in modeled_gids: 
            cell = pam.loc[cluster, accession]
            if type(cell) != float:  
                # get all the sequences not containing a premature stop:
                seqs = [i for i in cell.split(';') if i.endswith('_stop')==False]
                if len(seqs) >= 1:  
                    # populate the dict:
                    dups_dict[cluster] = seqs
    
    
    # iterate reactions:
    for r in ss_model.reactions:
        if 'Cluster_' in r.gene_reaction_rule: 
            gpr = r.gene_reaction_rule
            
            
            # force each gid to be surrounded by spaces: 
            gpr = ' ' + gpr.replace('(', ' ( ').replace(')', ' ) ') + ' '
            for cluster in dups_dict.keys():
                
                
                # search this gid surrounded by spaces:
                if f' {cluster} ' in gpr:
                    gpr = gpr.replace(f' {cluster} ', f' ({" or ".join(dups_dict[cluster])}) ')

            
            # remove spaces between parenthesis
            gpr = gpr.replace(' ( ', '(').replace(' ) ', ')')
            # remove spaces at the extremes: 
            gpr = gpr[1: -1]
            
            
            # New genes are introduced. Parethesis at the extremes are removed if not necessary. 
            r.gene_reaction_rule = gpr
            r.update_genes_from_gpr()
            
            
    # remaining old 'Cluster_'s need to removed.
    # remove if (1) hte ID starts with clusters AND (2) they are no more associated with any reaction
    to_remove = [g for g in ss_model.genes if (g.id.startswith("Cluster_") and len(g.reactions)==0)]
    cobra.manipulation.delete.remove_genes(ss_model, to_remove, remove_reactions=True)
    
    
    # check if the cleaning was successful
    for g in ss_model.genes:
        if g.id.startswith("Cluster_"):
            logger.error(f"Gene ID translation was not successful ({g.id} remained, accession {accession}). Please contact the developer.")
            return 1
        
        
    return 0



def annotate_genes(ss_model, accession, gannots):
    
    
    
    if gannots is not None: 
        # select the rows belonging to this accession: 
        gannots_accession = gannots[gannots['accession']==accession]
        # get the modeled gids: 
        gids_modeled = set([g.id for g in ss_model.genes])


        # iterate the 'gannots' dataframe:
        for index, row in gannots_accession.iterrows(): 
            if row['gid'] in gids_modeled:
                g = ss_model.genes.get_by_id(row['gid'])


                # '== str' to exclude empty cells: 
                if type(row['refseq']) == str:
                    g.annotation['refseq'] = [row['refseq']]
                if type(row['ncbiprotein']) == str:
                    g.annotation['ncbiprotein'] = [row['ncbiprotein']]
                if type(row['ncbigene']) == str:
                    g.annotation['ncbigene'] = [row['ncbigene']]
                if type(row['kegg']) == str:
                    g.annotation['kegg.genes'] = [row['kegg']]
                if type(row['uniprot']) == str:
                    g.annotation['uniprot'] = [row['uniprot']]
                
                
    # SBO annotation has been lost during the gene ID translation:
    for g in ss_model.genes:
        g.annotation['sbo'] = ['SBO:0000243']  # generic gene
                

def task_derivestrain(accession, args):
    
    
    # get the arguments
    panmodel = args['panmodel']
    pam = args['pam']
    report = args['report']
    outdir = args['outdir']
    gannots = args['gannots']
    logger = args['logger']
    sbml = args['sbml']
    
    
    # define key objects: 
    ss_model = panmodel.copy()  # create strain specific model
    
    
    # remove Clusters
    response = subtract_clusters(ss_model, panmodel, pam, accession)
    if response == 1: return 1   # will raise an error since a dict is expected
    
    # translate gene IDs:
    response = translate_genes(logger, ss_model, panmodel, pam, accession)
    if response == 1: return 1   # will raise an error since a dict is expected
    
    # annotate genes:
    response = annotate_genes(ss_model, accession, gannots)
    if response == 1: return 1   # will raise an error since a dict is expected
    
    
    # get the associated species:
    report = report[report['accession'] == accession]
    if len(report) > 1: 
        logger.error("Duplicated accessions in the provided report. Please report this error to the developer.")
    report = report.reset_index(drop=True)
    species = report.loc[0, 'species']
    
    
    # get some metrics: 
    n_G = len(ss_model.genes)
    n_R = len(ss_model.reactions)
    n_M = len(ss_model.metabolites)
    
    
    # try the FBA: 
    res = ss_model.optimize()
    obj_value = res.objective_value
    status = res.status
    
    
    # save strain specific model to disk
    cobra.io.save_json_model(ss_model, f'{outdir}/{accession}.json')
    if sbml: cobra.io.write_sbml_model(ss_model, f'{outdir}/{accession}.xml')
    
    
    # compose the new row:
    return [{'accession': accession, 'species': species, 'G': n_G, 'R': n_R, 'M': n_M, 'obj_value': obj_value, 'status': status }]



def  derive_strain_specific(logger, outdir, cores, panmodel, pam, report, gannots, sbml):

    
    # log some messages
    logger.info("Deriving strain-specific models...")
    
   
    # create output dir
    if os.path.exists(outdir + 'strain_models/'):
        # always overwriting if already existing
        shutil.rmtree(outdir + 'strain_models/')  
    os.makedirs(outdir + 'strain_models/', exist_ok=True)
    

    # create items for parallelization: 
    items = []
    for accession in pam.columns:
        items.append(accession)
        
        
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession', 'G', 'R', 'M', 'obj_value', 'status']), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_derivestrain),  # will return a new sequences dataframe (to be concat).
            itertools.repeat({'panmodel': panmodel, 'pam': pam, 'report': report, 'outdir': outdir + 'strain_models', 'gannots': gannots, 'logger': logger, 'sbml': sbml}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    # save tabular output:
    all_df_combined.to_csv(outdir + 'derive_strains.csv')
    
    
    return 0


import pickle
import os
import subprocess
import multiprocessing
import itertools
import shutil
import glob


import pandas as pnd
import seaborn as sb
import matplotlib.pyplot as plt
from matplotlib.patches import Patch


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results


            
def task_annotation(genome, args):
    
    
    # get the basename without extension:
    basename = os.path.basename(genome)
    accession, _ = os.path.splitext(basename)


    # launch the command
    with open(f'working/logs/stdout_annot_{accession}.txt', 'w') as stdout, open(f'working/logs/stderr_annot_{accession}.txt', 'w') as stderr: 
        command = f"""prokka --force --quiet \
            --cpus 1 \
            --outdir working/proteomes/ \
            --prefix {accession} \
            --noanno \
            --norrna \
            --notrna \
            {genome}"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
        
        
    # convert gff files to dataframe: 
    coords_df = []
    with open(f'working/proteomes/{accession}.gff', 'r') as r_handler: 
        file = r_handler.read()
        file = file.split('##FASTA', 1)[0] # not interested in contigs
        for line in file.split('\n'):
            if line.startswith('#'): continue
            if line=='': continue
            row = line.split('\t')
            row_dict = {key: value for key, value in zip(['contig', 'col2', 'col3', 'start', 'end', 'col6', 'strand', 'col8', 'attributes'], row)}
            # extract attributes from col9:
            for attribute in row_dict['attributes'].split(';'):
                attribute_id, attribute_value = attribute.split('=', 1)
                row_dict[attribute_id] = attribute_value
            # add the accession column:
            row_dict['accession'] = accession
            coords_df.append(row_dict)
    coords_df = pnd.DataFrame.from_records(coords_df)
    
    
    # format the dataframe and save
    coords_df = coords_df[['ID', 'accession', 'contig', 'strand', 'start', 'end']]
    coords_df.to_csv(f'working/coordinates/{accession}.csv')
        
        
    # remove useless files:
    os.remove(f'working/proteomes/{accession}.err')
    os.remove(f'working/proteomes/{accession}.ffn')
    os.remove(f'working/proteomes/{accession}.fna')
    os.remove(f'working/proteomes/{accession}.fsa')
    os.remove(f'working/proteomes/{accession}.gbk')
    os.remove(f'working/proteomes/{accession}.log')
    os.remove(f'working/proteomes/{accession}.sqn')
    os.remove(f'working/proteomes/{accession}.tbl')
    os.remove(f'working/proteomes/{accession}.tsv')
    os.remove(f'working/proteomes/{accession}.txt')
    # gff files could be useful for pangenome analysis like Raory/Panaroo:
    shutil.move(f'working/proteomes/{accession}.gff', f'working/gff/{accession}.gff')
        
    
    # return a row for the dataframe
    return [{'accession': accession, 'completed': True}]



def create_species_to_proteome(logger):
    
    
    # load the previously created species_to_genome: 
    with open('working/genomes/species_to_genome.pickle', 'rb') as handler:
        species_to_genome = pickle.load(handler)
        
        
    # create the species-to-proteome dictionary:
    species_to_proteome = {}
    for species in species_to_genome.keys(): 
        species_to_proteome[species] = []
        for genome in species_to_genome[species]: 
            basename = os.path.basename(genome)
            accession, _ = os.path.splitext(basename)
            species_to_proteome[species].append(f'working/proteomes/{accession}.faa')
    logger.debug(f"Created the species-to-proteome dictionary: " + str(species_to_proteome))
    
            
    # save the dictionary to disk: 
    with open('working/proteomes/species_to_proteome.pickle', 'wb') as file:
        pickle.dump(species_to_proteome, file)
    logger.debug(f"Saved the species-to-proteome dictionary to file: ./working/proteome/species_to_proteome.pickle.")
    
    
    
def create_seq_to_coords(logger):
    
    
    # load the previously created species_to_proteome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)

    
    # create a list of accessions to parse: 
    accessions = set()
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            accessions.add(accession)
            
    
    # quick check of the available dict to save time:
    if os.path.exists(f'working/coordinates/seq_to_coords.pickle'):
        accessions_available = set()
        with open('working/coordinates/seq_to_coords.pickle', 'rb') as handler:
            seq_to_coords = pickle.load(handler)
        for attribs in seq_to_coords.values():
            accessions_available.add(attribs['accession'])
        if accessions == accessions_available:
            logger.debug(f"Found a good sequence-to-coordinates dictionary file already available.")
            return
    
        
    # create the dict seq-to-coords
    seq_to_coords = {}
    for accession in accessions: 
        coords_df = pnd.read_csv(f'working/coordinates/{accession}.csv')
        for index, row in coords_df.iterrows(): 
            if type(row['ID']) != str:
                continue
            # 'str' needed as some accessions could be just numbers
            seq_to_coords[row['ID']] = {'accession': str(row['accession']), 'contig': str(row['contig']), 'strand': str(row['strand']), 'start': row['start'], 'end': row['end']}
    
    
    # save the dictionary to disk: 
    with open('working/coordinates/seq_to_coords.pickle', 'wb') as file:
        pickle.dump(seq_to_coords, file)
    logger.debug(f"Saved the sequence-to-coordinates dictionary to file: ./working/coordinates/seq_to_coords.pickle.")
    


def figure_cds(logger, outdir):
    
    
    logger.info("Producing figure for extracted CDSs in {outdir}/figures/n_cds.png...")
    
    # create summary dataframe  (should be equivalent to 'all_df_combined'): 
    prodigal_summary = []
    for file in glob.glob(f"working/proteomes/*.faa"):
        accession = file.replace('working/proteomes/', '').replace('.faa', '')
        n_cds = open(file).read().count('>')
        prodigal_summary.append({'assembly_accession': accession, 'n_cds': n_cds})
    prodigal_summary = pnd.DataFrame.from_records(prodigal_summary)
    prodigal_summary = prodigal_summary.set_index('assembly_accession', drop=True, verify_integrity=True)
    
    # load the genomes_df to have the 'strain_isolate' and 'organism_name' columns:
    genomes_df = pnd.read_csv('working/genomes/genomes.csv', index_col=0)
    genomes_df = genomes_df.set_index('assembly_accession', drop=True, verify_integrity=True)
    
    # concat the dataframes:
    df = pnd.concat([genomes_df, prodigal_summary], axis=1)
    
    
    # define colors:
    df = df.set_index('strain_isolate', drop=False)
    colors = df['organism_name'].map({species: f'C{number}' for number, species in enumerate(df['organism_name'].unique())}).to_dict()
        
    # draw bars:
    fig, ax = plt.subplots()
    _ = sb.barplot(df, x='strain_isolate', y='n_cds', palette=colors, hue='strain_isolate', legend=False, ax=ax)
    
    # set tick labels
    ax.tick_params(axis='x', labelrotation=90)
    [label.set_color(colors[label.get_text()]) for label in ax.get_xticklabels()]
    
    # set legend:
    plt.legend(handles=[Patch(color=f'C{number}', label=species) for number, species in enumerate(df['organism_name'].unique())], title='', loc='center left', bbox_to_anchor=(1.05, 0.5))
    
    ax.figure.set_size_inches(0.2*len(df), 4)
    sb.despine()

    try: 
        plt.savefig(outdir + 'figures/n_cds.png', dpi=300, bbox_inches='tight')
    except:  # the png image could be too large, so we produce svg
        logger.info(f"PNG was too large: producing the SVG version instead {outdir}/figures/n_cds.svg...")
        plt.savefig(outdir + 'figures/n_cds.svg', bbox_inches='tight')
        
    
    
    
def extract_cds(logger, cores, outdir, nofig):
    
    
    # create sub-directory without overwriting:
    logger.info("Extracting the CDSs from the genomes...")
    os.makedirs('working/proteomes/', exist_ok=True)
    os.makedirs('working/coordinates/', exist_ok=True)
    os.makedirs('working/gff/', exist_ok=True)
    os.makedirs(outdir + 'figures/', exist_ok=True)


    # load the previously created species_to_genome: 
    with open('working/genomes/species_to_genome.pickle', 'rb') as handler:
        species_to_genome = pickle.load(handler)


    # create items for parallelization: 
    items = []
    for species in species_to_genome.keys(): 
        for genome in species_to_genome[species]: 
            items.append(genome)
            
            
    # check if the corresponding proteomes are already available: 
    already_computed = []
    for genome in items: 
        basename = os.path.basename(genome)
        accession, _ = os.path.splitext(basename)
        already_computed.append(os.path.exists(f'working/proteomes/{accession}.faa'))
        already_computed.append(os.path.exists(f'working/coordinates/{accession}.csv'))
    if all(already_computed):
        logger.info("Found all the needed files already computed. Skipping this step.")
        # save the species_to_proteome and seq_to_coords dicts
        create_species_to_proteome(logger)
        create_seq_to_coords(logger)
        
        if not nofig:
            figure_cds(logger, outdir)
        return 0
    

    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession', 'completed']), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_annotation),
            itertools.repeat({}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)  # all_df_combined can be ignored.
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join()  
    
    
    # save the species_to_proteome and seq_to_coords dicts
    create_species_to_proteome(logger)
    create_seq_to_coords(logger)
    
    
    if not nofig:
        figure_cds(logger, outdir)
    return 0



def handle_manual_proteomes(logger, proteomes):
    
    
    # create a species-to-genome dictionary
    species_to_proteome = {}
    logger.debug(f"Checking the formatting of the provided -p/--proteomes attribute...") 
    
    
    # check if the user specified a folder:
    if os.path.exists(proteomes) and os.path.isdir(proteomes):
        if proteomes[-1] != '/': proteomes = proteomes + '/'
        files = glob.glob(proteomes + '*')
        species_to_proteome['Spp'] = files
    
    elif '+' in proteomes and '@' in proteomes: 
        for species_block in proteomes.split('+'):
            species, files = species_block.split('@')
            for file in files.split(','): 
                if not os.path.exists(file):
                    logger.error("The following file provided in -p/--proteomes does not exists: " + file)
                    return 1
            species_to_proteome[species] = files.split(',')
            
    else: # the user has just 1 species
        for file in proteomes.split(','): 
            if not os.path.exists(file):
                logger.error("The following file provided in -p/--proteomes does not exists: " + file)
                return 1
        species_to_proteome['Spp'] = proteomes.split(',')

    
    # report a summary of the parsing: 
    logger.info(f"Inputted {len(species_to_proteome.keys())} species with well-formatted paths to proteomes.") 
    
    
    # move the genomes to the usual directory: 
    os.makedirs('working/proteomes/', exist_ok=True)
    for species in species_to_proteome.keys():
        copied_files = []
        for file in species_to_proteome[species]:
            basename = os.path.basename(file)
            shutil.copyfile(file, 'working/proteomes/' + basename)
            copied_files.append('working/proteomes/' + basename)
        species_to_proteome[species] = copied_files
    logger.debug(f"Input proteomes copied to ./working/proteomes/.")
    logger.debug(f"Created the species-to-proteome dictionary: {str(species_to_proteome)}.") 
    
    
    # save the dictionary to disk: 
    with open('working/proteomes/species_to_proteome.pickle', 'wb') as file:
        pickle.dump(species_to_proteome, file)
    logger.debug(f"Saved the species-to-proteome dictionary to file: ./working/proteomes/species_to_proteome.pickle.")
    
    
    return 0






import pickle
import os
import subprocess
import multiprocessing
import itertools
import glob
import json
import shutil 


import pandas as pnd
import seaborn as sb
import matplotlib.pyplot as plt
from matplotlib.patches import Patch


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import get_allmeta_df



def task_bmetrics(proteome, args): 
        
        
    # retrive the arguments:
    buscodb = args['buscodb']


    # get the basename without extension:
    basename = os.path.basename(proteome)
    accession, _ = os.path.splitext(basename)


    # launch the command
    with open(f'working/logs/stdout_bmetrics_{accession}.txt', 'w') as stdout, open(f'working/logs/stderr_bmetrics_{accession}.txt', 'w') as stderr: 
        command = f"""busco -f --cpu 1 --offline \
            -i {proteome} \
            --mode proteins \
            --lineage_dataset {buscodb} \
            --download_path working/bmetrics/db/ \
            --out_path working/bmetrics/ \
            --out {accession}"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()


    # return a row for the dataframe
    return [{'accession': accession, 'completed': True}]



def compute_bmetrics(logger, cores, buscodb): 
    
    
    # logger message
    logger.info("Calculating the biological metrics to filter the genomes...")
    
    
    # load the previously created species_to_proteome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        
    
    # check if the metrics were already computed: 
    if os.path.exists('working/filtering/bmetrics.csv'):
        bmetrics_df = pnd.read_csv('working/filtering/bmetrics.csv', index_col=0)
        presence_list = []
        for species in species_to_proteome.keys(): 
            for proteome in species_to_proteome[species]:
                # get the basename without extension:
                basename = os.path.basename(proteome)
                accession, _ = os.path.splitext(basename)
                presence_list.append(accession in bmetrics_df['accession'].to_list())
        if all(presence_list): 
            logger.info("Found all the needed files already computed. Skipping this step.")
            return 0
    
    
    # create the worlder for biological metrics: 
    os.makedirs('working/bmetrics/', exist_ok=True)
    
    
    # check if the user specified a database
    if buscodb == 'bacteria_odb10': 
        logger.warning("We strongly suggest to set a more specific Busco database instead of 'bacteria_odb10' (use -b/--buscodb). To show the available Busco databases type 'gempipe recon -b show'.")
    
    
    # assuring the presence of the specified database
    logger.debug("Downloading the specified BUSCO database...")
    with open(f'working/logs/stdout_bmetrics_dbdownload.txt', 'w') as stdout, open(f'working/logs/stderr_bmetrics_dbdownload.txt', 'w') as stderr: 
        command = f"""busco -f \
            --download_path working/bmetrics/db/ \
            --out_path working/bmetrics/ \
            --download {buscodb}"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
    logger.debug(f"Download completed for {buscodb}.")
        
      
    # create items for parallelization: 
    items = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]: 
            items.append(proteome)
            
            
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession', 'completed']), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_bmetrics),
            itertools.repeat({'buscodb': buscodb}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)  # all_df_combined can be ignored.
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
            
    
    # logger message
    logger.debug("Finished to compute the biological metrics.")
    
    
    # collect the short summaries:
    logger.debug("Gathering the short summaries...")
    bmetrics_df = []  
    for file in glob.glob(f"working/bmetrics/*/run_{buscodb}/short_summary.json"): 
        accession = file.replace('working/bmetrics/', '')
        accession = accession.replace(f'/run_{buscodb}/short_summary.json', '')
        jsonout = json.load(open(file, 'r'))
        
        
        # handle different versions of busco
        try: C = jsonout['results']['Complete']
        except: C = jsonout['results']['Complete percentage']
        try: sC = jsonout['results']['Single copy']
        except: sC = jsonout['results']['Single copy percentage']
        try: mC = jsonout['results']['Multi copy']
        except: mC = jsonout['results']['Multi copy percentage']
        try: F = jsonout['results']['Fragmented']
        except: F = jsonout['results']['Fragmented percentage']
        try: M = jsonout['results']['Missing']
        except: M = jsonout['results']['Missing percentage']
        
        
        bmetrics_df.append({
            'accession': accession,
            'C': C,
            'Single copy': sC,
            'Multi copy': mC,
            'F': F,
            'M': M,
            'n_markers': jsonout['results']['n_markers']
        })
    bmetrics_df = pnd.DataFrame.from_records(bmetrics_df)
    os.makedirs('working/filtering/', exist_ok=True)
    bmetrics_df.to_csv('working/filtering/bmetrics.csv')
    logger.debug("Biological metrics saved to ./working/filtering/bmetrics.csv.")
    
    
    # cleaning the workspace from useless files: 
    shutil.rmtree('working/bmetrics/')
    for file in glob.glob('./busco_*.log'): os.remove(file)
    logger.debug("Removed useless files.")
    
    
    return 0



def compute_tmetrics(logger, cores):
    
    
    # logger message
    logger.info("Calculating the technical metrics to filter the genomes...")
    
    
    # load the previously created species_to_genome: 
    with open('working/genomes/species_to_genome.pickle', 'rb') as handler:
        species_to_genome = pickle.load(handler)
        
        
    # check if the metrics were already computed: 
    if os.path.exists('working/filtering/tmetrics.csv'):
        tmetrics_df = pnd.read_csv('working/filtering/tmetrics.csv', index_col=0)
        presence_list = []
        for species in species_to_genome.keys(): 
            for genome in species_to_genome[species]:
                # get the basename without extension:
                basename = os.path.basename(genome)
                accession, _ = os.path.splitext(basename)
                presence_list.append(accession in tmetrics_df['accession'].to_list())
        if all(presence_list): 
            logger.info("Found all the needed files already computed. Skipping this step.")
            return 0

        
    # create the list of genomes to evaluate: 
    genome_files = []
    for species in species_to_genome.keys(): 
        for genome in species_to_genome[species]: 
            genome_files.append(genome)

            
    # launch the command
    with open(f'working/logs/stdout_tmetrics.txt', 'w') as stdout, open(f'working/logs/stderr_tmetrics.txt', 'w') as stderr: 
        command = f"""seqkit stats \
            --tabular \
            --basename \
            --all \
            --threads {cores} \
            --out-file working/filtering/tmetrics.csv \
            {' '.join(genome_files)}"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
        
        
    # format the table:
    tmetrics_df = pnd.read_csv('working/filtering/tmetrics.csv', sep='\t')
    tmetrics_df = tmetrics_df.rename(columns={'file': 'accession', 'num_seqs': 'ncontigs'})
    tmetrics_df['accession'] = tmetrics_df['accession'].apply(lambda x: os.path.splitext(x)[0])
    tmetrics_df.to_csv('working/filtering/tmetrics.csv')
    logger.debug("Technical metrics saved to ./working/filtering/tmetrics.csv.")
                
                
    # logger message:
    logger.debug(f"Finished to compute the technical metrics.")
    
    
    return 0
    
    
    
def figure_bmetrics(logger, outdir, bad_genomes): 
    
    
    logger.info("Producing figure for biological metrics in {outdir}/figures/busco.png...")
    
    df = get_allmeta_df()

    # create new col to show filtering: 
    df['excluded'] = 0
    for accession, row in df.iterrows(): 
        if accession in bad_genomes: 
            df.loc[accession, 'excluded'] = 100
    
    # define colors:
    df = df.set_index('strain_isolate', drop=False)
    colors = df['organism_name'].map({species: f'C{number}' for number, species in enumerate(df['organism_name'].unique())}).to_dict()
    
    # draw bars:
    fig, ax = plt.subplots()
    _ = sb.barplot(df, x='strain_isolate', y='C', color='C8', ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='F', color='C9', bottom=df['C'], ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='M', color='C4', bottom=df['C']+df['F'], ax=ax)
    _ = sb.barplot(df, x='strain_isolate', y='excluded', color='white', alpha=0.55, ax=ax)
    
    # set tick labels:
    ax.tick_params(axis='x', labelrotation=90)
    [label.set_color(colors[label.get_text()]) for label in ax.get_xticklabels()]
    
    # set legends:
    l1 = plt.legend(handles=[Patch(color=color, label=metric) for color, metric in zip(['C8','C9','C4'], ['BUSCO C','BUSCO F','BUSCO M'])], title='', loc='upper left', bbox_to_anchor=(1.05, 0.5))
    l2 = plt.legend(handles=[Patch(color=f'C{number}', label=species) for number, species in enumerate(df['organism_name'].unique())], title='', loc='lower left', bbox_to_anchor=(1.05, 0.5))
    ax.add_artist(l1)  # l2 implicitly replaces l1
    
    ax.figure.set_size_inches(0.2*len(df), 4)
    ax.set_ylabel('%')
    sb.despine()
    
    try: 
        plt.savefig(outdir + 'figures/bmetrics.png', dpi=300, bbox_inches='tight')
    except:  # the png image could be too large, so we produce svg
        logger.info(f"PNG was too large: producing the SVG version instead {outdir}/figures/bmetrics.svg...")
        plt.savefig(outdir + 'figures/bmetrics.svg', bbox_inches='tight')
        


def figure_tmetrics(logger, outdir, bad_genomes): 
    
    
    for tmetric in ['ncontigs', 'N50', 'sum_len']: 
        logger.info(f"Producing figure for technical metric {tmetric} in {{outdir}}/figures/{tmetric}.png...")
    
    
        df = get_allmeta_df()
        df['sum_len'] = df['sum_len'].apply(lambda x: x / 1000 / 1000)  # convert to Mb
        
        # create new col to show filtering: 
        df['excluded'] = 0
        for accession, row in df.iterrows(): 
            if accession in bad_genomes: 
                df.loc[accession, 'excluded'] = df.loc[accession, tmetric]

        # define colors:
        df = df.set_index('strain_isolate', drop=False)
        colors = df['organism_name'].map({species: f'C{number}' for number, species in enumerate(df['organism_name'].unique())}).to_dict()

        # draw bars:
        fig, ax = plt.subplots()
        _ = sb.barplot(df, x='strain_isolate', y=tmetric, hue='strain_isolate', legend=False, palette=colors, ax=ax)
        _ = sb.barplot(df, x='strain_isolate', y='excluded', color='white', alpha=0.55, ax=ax)

        # set tick labels:
        ax.tick_params(axis='x', labelrotation=90)
        [label.set_color(colors[label.get_text()]) for label in ax.get_xticklabels()]

        # set legend:
        plt.legend(handles=[Patch(color=f'C{number}', label=species) for number, species in enumerate(df['organism_name'].unique())], title='', loc='center left', bbox_to_anchor=(1.05, 0.5))

        ax.figure.set_size_inches(0.2*len(df), 4)
        ax.set_ylabel(tmetric)
        sb.despine()

        try: 
            plt.savefig(outdir + f'figures/{tmetric}.png', dpi=300, bbox_inches='tight')
        except:  # the png image could be too large, so we produce svg
            logger.info(f"PNG was too large: producing the SVG version instead {outdir}/figures/{tmetric}.svg...")
            plt.savefig(outdir + f'figures/{tmetric}.svg', bbox_inches='tight')
            


def filter_genomes(logger, cores, buscodb, buscoM, buscoF, ncontigs, N50, outdir, nofig):
    
    
    # compoute biological metrics: 
    response = compute_bmetrics(logger, cores, buscodb)
    if response == 1: return 1
    
    
    # compute technical metrics: 
    response = compute_tmetrics(logger, cores)
    if response == 1: return 1
    
    
    # read the metrics tables
    bmetrics_df = pnd.read_csv('working/filtering/bmetrics.csv', index_col=0)
    tmetrics_df = pnd.read_csv('working/filtering/tmetrics.csv', index_col=0)
    
    
    # get the number of Busco's scingle-copy orthologs: 
    n_sco = list(set(bmetrics_df['n_markers'].to_list()))[0]
    
    
    # check the inputted buscoM / buscoF
    busco_metrics = {'buscoM': buscoM, 'buscoF': buscoF}
    for metric, value in busco_metrics.items():
        if value.endswith('%'): 
            value = value[:-1]
            value = int(value)
        else: 
            value = int(value)
            value = value / n_sco * 100
        busco_metrics[metric] = value
        
        
    # filter genomes and proteomes based on metrics
    all_genomes = set(tmetrics_df['accession'].to_list())
    bmetrics_good = set(bmetrics_df[(bmetrics_df['M'] <= busco_metrics['buscoM']) & (bmetrics_df['F'] <= busco_metrics['buscoF'])]['accession'].to_list())
    tmetrics_good = set(tmetrics_df[(tmetrics_df['ncontigs'] <= ncontigs) & (tmetrics_df['N50'] >= N50)]['accession'].to_list())
    good_genomes = bmetrics_good.intersection(tmetrics_good)
    bad_genomes = all_genomes - good_genomes
        
    
    # load the previously created dictionaries: 
    with open('working/genomes/species_to_genome.pickle', 'rb') as handler:
        species_to_genome = pickle.load(handler)
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
    
    
    # re-writes the dictionaries:
    species_to_genome_new = {}
    species_to_proteome_new = {}
    bad_quality = []
    for species in species_to_genome.keys(): 
        species_to_genome_new[species] = []
        species_to_proteome_new[species] = []
        for genome, proteome in zip(species_to_genome[species], species_to_proteome[species]):
            basename = os.path.basename(genome)
            accession, _ = os.path.splitext(basename)
            if accession in good_genomes:
                species_to_genome_new[species].append(genome)
                species_to_proteome_new[species].append(proteome)
            else: 
                logger.debug("Found a bad quality genome: " + genome + ". Will be ignored in subsequent analysis.")
                bad_quality.append(genome)
    logger.info(f"Found {len(bad_quality)} bad quality genomes. They will be ignored in subsequent analysis. Use --verbose to see the list.")
                
    with open('working/genomes/species_to_genome.pickle', 'wb') as file:
        pickle.dump(species_to_genome_new, file)
    with open('working/proteomes/species_to_proteome.pickle', 'wb') as file:
        pickle.dump(species_to_proteome_new, file)
        
        
    # produce plot for genome filtering 
    if not nofig: 
        os.makedirs(outdir + 'figures/', exist_ok=True)
        figure_bmetrics(logger, outdir, bad_genomes)
        figure_tmetrics(logger, outdir, bad_genomes)
    
    
    return 0
            
        
    
    


import argparse
import sys
import multiprocessing 
import logging 
from logging.handlers import QueueHandler
import traceback
import importlib.metadata
from datetime import datetime


from .recon import recon_command
from .derive import derive_command
from .autopilot import autopilot_command


from .curate.gaps import *
from .curate.sanity import *
from .curate.medium import *
# set up the cobra solver


# cobra was already imported from other statements above
# set the global solver:
#try: cobra.Configuration().solver = "cplex"
#except:  cobra.Configuration().solver = "glpk" # "glpk_exact"
# get the global solver:
cobra_config = cobra.Configuration()
solver_name = str(cobra_config.solver.log).split(' ')[1]
solver_name = solver_name.replace("optlang.", '')
solver_name = solver_name.replace("_interface", '')


from .flowchart import Flowchart



def main(): 
    
    
    # define the header of main- and sub-commands. 
    pub_details = 'TODO'
    header = f'gempipe v{importlib.metadata.metadata("gempipe")["Version"]}, please cite "{pub_details}".\nFull documentation available at https://gempipe.readthedocs.io/en/latest/index.html.'
    
    
    # create the command line arguments:
    parser = argparse.ArgumentParser(description=header, add_help=False)
    parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gempipe')['Version']}", help="Show version number and exit.")
    subparsers = parser.add_subparsers(title='gempipe subcommands', dest='subcommand', help='', required=True)

    
    # create the 3 subparsers:
    recon_parser = subparsers.add_parser('recon', description=header, help='Reconstruct a draft pan-model and a PAM.', formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    derive_parser = subparsers.add_parser('derive', description=header, help='Derive strain- and species-specific models.', formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    autopilot_parser = subparsers.add_parser('autopilot', description=header, help='Run recon + derive, with automated pan-model gap-filling. Use with consciousness!', formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    
    
    # add arguments for the 'derive' command
    derive_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    derive_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gempipe')['Version']}", help="Show version number and exit.")
    derive_parser.add_argument("-c", "--cores", metavar='', type=int, default=1, help="How many parallel processes to use.")
    derive_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    derive_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    derive_parser.add_argument("-im", "--inpanmodel", metavar='', type=str, default='-', help="Path to the input pan-model.")
    derive_parser.add_argument("-ip", "--inpam", metavar='', type=str, default='-', help="Path to the input PAM.")
    derive_parser.add_argument("-ir", "--inreport", metavar='', type=str, default='-', help="Path to the input report file.")
    derive_parser.add_argument("-ig", "--ingannots", metavar='', type=str, default='-', help="Path to the input genes annotation file.")
    derive_parser.add_argument("-m", "--media", metavar='', type=str, default='-', help="Medium definition file or folder containing media definitions, to be used during the automatic gap-filling.")
    derive_parser.add_argument("--minflux", metavar='', type=float, default=0.1, help="Minimum flux through the objective of strain-specific models.")
    derive_parser.add_argument("--biolog", action='store_true', help="Simulate Biolog's utilization tests on strain-specific models.")
    derive_parser.add_argument("--sbml", action='store_true', help="Save the output GSMMs in SBML format (L3V1 FBC2) in addition to JSON.")
    derive_parser.add_argument("--skipgf", action='store_true', help="Skip the gap-filling step applied to the strain-specific models.")
    
    
    # add arguments for the 'recon'/'autopilot' command
    for subparser in [recon_parser, autopilot_parser]:
        subparser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
        subparser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gempipe')['Version']}", help="Show version number and exit.")
        subparser.add_argument("-c", "--cores", metavar='', type=int, default=1, help="Number of parallel processes to use.")
        subparser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
        subparser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
        subparser.add_argument("--overwrite", action='store_true', help="Delete the working/ directory at the startup.")
        subparser.add_argument("--dbs", metavar='', type=str, default='./working/dbs/', help="Path were the needed databases are stored (or downloaded if not already existing).")
        subparser.add_argument("-t", "--taxids", metavar='', type=str, default='-', help="Taxids of the species to model (comma separated, for example '252393,68334').")
        subparser.add_argument("-g", "--genomes", metavar='', type=str, default='-', help="Input genome files or folder containing the genomes (see documentation).")
        subparser.add_argument("-p", "--proteomes", metavar='', type=str, default='-', help="Input proteome files or folder containing the proteomes (see documentation).")
        subparser.add_argument("-gb", "--genbanks", metavar='', type=str, default='-', help="Input genbank files (.gb, .gbff) or folder containing the genbanks (see documentation).")
        subparser.add_argument("-s", "--staining", metavar='', type=str, default='neg', help="Gram staining, 'pos' or 'neg'.")
        subparser.add_argument("-b", "--buscodb", metavar='', type=str, default='bacteria_odb10', help="Busco database to use ('show' to see the list of available databases).")
        subparser.add_argument("--buscoM", metavar='', type=str, default='2%', help="Maximum number of missing Busco's single copy orthologs (absolute or percentage).")
        subparser.add_argument("--buscoF", metavar='', type=str, default='100%', help="Maximum number of fragmented Busco's single copy orthologs (absolute or percentage).")
        subparser.add_argument("--ncontigs", metavar='', type=int, default=200, help="Maximum number of contigs allowed per genome.")
        subparser.add_argument("--N50", metavar='', type=int, default=50000, help="Minimum N50 allowed per genome.")
        subparser.add_argument("--identity", metavar='', type=int, default=30, help="Minimum percentage amino acidic sequence identity to use when aligning against the BiGG gene database.")
        subparser.add_argument("--coverage", metavar='', type=int, default=70, help="Minimum percentage coverage to use when aligning against the BiGG gene database.")
        subparser.add_argument("-rm", "--refmodel", metavar='', type=str, default='-', help="Model to be used as reference.")
        subparser.add_argument("-rp", "--refproteome", metavar='', type=str, default='-', help="Proteome to be used as reference.")
        subparser.add_argument("-rs", "--refspont", metavar='', type=str, default='spontaneous', help="Reference gene marking spontaneous reactions.")
        subparser.add_argument("-mc", "--mancor", metavar='', type=str, default='-', help="Manual corrections to apply during the reference expansion.")
        subparser.add_argument("--tcdb", action='store_true', help="Experimental feature: try to build transport reactions using TCDB.")
        subparser.add_argument("--dedup", action='store_true', help="Try to remove duplicate metabolites and reactions using MNX annotation, when a reference is provided.")
        subparser.add_argument("--norec", action='store_true', help="Skip gene recovery when starting from genomes.")
        subparser.add_argument("--dbmem", action='store_true', help="Load the entire eggNOG-mapper database into memory (should speed up the functional annotation step).")
        subparser.add_argument("--sbml", action='store_true', help="Save the output GSMMs in SBML format (L3V1 FBC2) in addition to JSON.")
        subparser.add_argument("--nofig", action='store_true', help="Skip the generation of figures.")
        
    
    # add arguments specifically for the 'autopilot' command
    autopilot_parser.add_argument("-m", "--media", metavar='', type=str, default='-', help="Medium definition file or folder containing media definitions, to be used during the automatic gap-filling.")
    autopilot_parser.add_argument("--minflux", metavar='', type=float, default=0.1, help="Minimum flux through the objective of strain-specific models.")
    autopilot_parser.add_argument("--minpanflux", metavar='', type=float, default=0.3, help="Minimum flux through the objective of the pan model.")
    autopilot_parser.add_argument("--biolog", action='store_true', help="Simulate Biolog's utilization tests on strain-specific models.")
    

    # check the inputted subcommand, automatic sys.exit(1) if a bad subprogram was specied. 
    args = parser.parse_args()
    
    
    # set the multiprocessing context
    multiprocessing.set_start_method('fork') 
    
    
    # create a logging queue in a dedicated process.
    def logger_process_target(queue):
        logger = logging.getLogger('gempipe')
        while True:
            message = queue.get() # block until a new message arrives
            if message is None: # sentinel message to exit the loop
                break
            logger.handle(message)
    queue = multiprocessing.Queue()
    logger_process = multiprocessing.Process(target=logger_process_target, args=(queue,))
    logger_process.start()
    
    
    # connect the logger for this (main) process: 
    logger = logging.getLogger('gempipe')
    logger.addHandler(QueueHandler(queue))
    if args.verbose: logger.setLevel(logging.DEBUG) # debug (lvl 10) and up
    else: logger.setLevel(logging.INFO) # debug (lvl 20) and up
    
    
    # handy function to print without time/level (for header / trailer)
    def set_header_trailer_formatter(logger):
        formatter = logging.Formatter('%(message)s')
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return handler
    
    
    # to print the main pipeline logging:
    def set_usual_formatter(logger):
        formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s', datefmt="%H:%M:%S")
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return handler
    
    
    
    # show a welcome message:
    thf_handler = set_header_trailer_formatter(logger)
    logger.info(header + '\n')
    command_line = '' # print the full command line:
    for arg, value in vars(args).items():
        if arg == 'subcommand': command_line = command_line + f"gempipe {value} "
        else: command_line = command_line + f"--{arg} {value} "
    logger.info('Inputted command line: "' + command_line.rstrip() + '".\n')
    logger.removeHandler(thf_handler)
    
    
    
    usual_handler = set_usual_formatter(logger)
    current_date_time = datetime.now()
    formatted_date = current_date_time.strftime("%Y-%m-%d")
    logger.info(f"Welcome to gempipe! Launching the pipeline on {formatted_date}...")
    logger.info(f'COBRApy started with solver: {solver_name}.')
    try: 
        # choose which subcommand to lauch: 
        if args.subcommand == 'recon':
            response = recon_command(args, logger)
        if args.subcommand == 'derive':
            response = derive_command(args, logger)
        if args.subcommand == 'autopilot':
            response = autopilot_command(args, logger)
            
        if response == 0:
            logger.info("gempipe terminated without errors!")
    except: 
        # show the error stack trace for this un-handled error: 
        response = 1
        logger.error(traceback.format_exc())
    logger.removeHandler(usual_handler)


    
    # Terminate the program:
    thf_handler = set_header_trailer_formatter(logger)
    if response == 1: 
        queue.put(None) # send the sentinel message
        logger_process.join() # wait for all logs to be digested
        sys.exit(1)
    else: 
        # show a bye message
        queue.put(None) # send the sentinel message
        logger_process.join() # wait for all logs to be digested
        logger.info('\n' + header)
        sys.exit(0) # exit without errors
        
        
        
if __name__ == "__main__":
    main()
    



import pickle


import pandas as pnd
import cobra


from ..recon import preliminary_checks
from ..recon import draft_reconstruction
from ..recon import automated_curation


from .priogapfiller import prio_gapfiller


from ..derive import derive_all


from ..commons import get_md5_string
from ..commons import get_outdir



def autopilot_command(args, logger):
        
    
    response = preliminary_checks(args, logger)
    if response == 0: return 0
    if response == 1: return 1
    
    
    response = draft_reconstruction(args, logger)
    if response == 1: return 1


    # insert the prioritized gapfiller, just before the deduplication.
    # this will OVERWRITE the panmodel in 'working/duplicates/draft_panmodel.json'
    # taht was created during draft_reconstruction()!  
    response = prio_gapfiller(logger, args.refmodel, args.refproteome, args.staining, args.mancor, args.media, args.minpanflux)
    if response == 1: return 1


    # save the panmodel md5:
    panmodel_md5 = get_md5_string('working/duplicates/draft_panmodel.json')
    with open('working/duplicates/md5.pickle', 'wb') as handle: 
        pickle.dump(panmodel_md5, handle)

        
    response = automated_curation(args, logger)
    if response == 1: return 1


    # make a SBML copy of the final draft panmodel if requested
    if args.sbml: 
        outdir = get_outdir(args.outdir)
        # load the final draft panmodel (prio-gapfilled)
        draft_panmodel = cobra.io.load_json_model(outdir + 'draft_panmodel.json')
        # create a SBML copy
        cobra.io.write_sbml_model(draft_panmodel, outdir + 'draft_panmodel.xml')


    # now we have an already gap-filled and (partially) curated panmodel.
    # apply the derivation of strain- and species-specific metabolic models:
    # load input files:
    outdir = get_outdir(args.outdir)
    panmodel = cobra.io.load_json_model( outdir + 'draft_panmodel.json')
    pam = pnd.read_csv(outdir + 'pam.csv', index_col=0)
    report = pnd.read_csv(outdir + 'report.csv', index_col=0)
    
    if args.genbanks != '-':  # if the user provided genbanks, then there must but a 'gsnnots.csv' in output.
        # "low_memory=False" to prevent "DtypeWarning: Columns ... have mixed types."
        gannots = pnd.read_csv(outdir + 'gannots.csv', index_col=0, low_memory=False)
        gannots = gannots.astype(str)  # 'ncbigene' is still formatted as number
        gannots = gannots.replace('nan', None)
        #gannots['refseq'] = [i.rsplit('.', 1)[0] if type(i)==str else None for i in gannots['refseq'] ]
        gannots['ncbigene'] = [i.rsplit('.', 1)[0] if type(i)==str else None for i in gannots['ncbigene'] ]
    else: gannots = None

    
    response = derive_all(logger, outdir, args.cores, panmodel, pam, report, gannots, args.media, args.minflux, args.biolog, args.sbml, skipgf=False)
    if response == 1: return 1
    
    
    return 0



import uuid


from IPython.display import display, HTML



class Flowchart:
    
    
    
    def __init__(self, diagram):
        
        self.diagram = self.process_diagram(diagram)
        self.uid = uuid.uuid4()

        
        
    def process_diagram(self, diagram):
        
        diagram = diagram.replace("\n", "\\n")
        diagram = diagram.lstrip("\\n")
        diagram = diagram.replace("'", '"')
        return diagram

    
    
    def render(self, height=500, zoom=1):
        
        html = f"""
        <style> #outcellbox {{display: flex; justify-content: center; overflow: hidden; width: 99%; height: {height}px; background-color: #ffffff; border: 1px solid grey;}} </style>
        
        <script src="https://unpkg.com/@panzoom/panzoom@4.5.1/dist/panzoom.min.js"></script>

        <div class="mermaid-{self.uid}" id="outcellbox"></div> 
        <script type="module">
            import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10.6.1/+esm';
            const graphDefinition = \'___diagram___\';
            const element = document.querySelector('.mermaid-{self.uid}');
            const {{ svg }} = await mermaid.render('graphDiv-{self.uid}', graphDefinition);
            element.innerHTML = svg;
            
            const elem = document.getElementById('graphDiv-{self.uid}');
            const panzoom = Panzoom(elem, {{maxScale: 50}});
            panzoom.zoom({zoom});
            elem.parentElement.addEventListener('wheel', panzoom.zoomWithWheel);
        </script>
        """
        html = html.replace("___diagram___", self.diagram)
        return display(HTML(html))


import multiprocessing
import itertools
import os
import glob
import shutil


import pandas as pnd
import cobra


from gempipe.curate.gaps import perform_gapfilling
from gempipe.curate.gaps import import_from_universe
from gempipe.curate.gaps import get_solver


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import get_media_definitions
from ..commons import apply_json_medium
from ..commons import check_panmodel_growth
from ..commons import strenghten_uptakes



def task_strainfiller(file, args):
    
    
    # get the arguments
    panmodel = args['panmodel']
    minflux = args['minflux']
    outdir = args['outdir']
    media = args['media']
    sbml = args['sbml']
    
    
    # load the strain-specific model
    ss_model = cobra.io.load_json_model(file)

    
    # get the accession
    basename = os.path.basename(file)
    accession, _ = os.path.splitext(basename)
    
    
    # define key objects:
    gapfilling_failed = False
    inserted_rids_medium = {}  # track the inserted rids for each provided medium
    obj_value_medium = {}
    status_medium = {}
    
    
    # iterate the provided media
    for medium_name, medium in media.items():


        # apply the medium recipe:
        # response should be alwys 0 as the growth of panmodel was previously assessed. 
        response = apply_json_medium(panmodel, medium)
        response = apply_json_medium(ss_model, medium)
        
    
        # first try of the gapfilling algo: 
        inserted_rids = []  # rids needed for gapfilling
        first_sol_rids = perform_gapfilling(ss_model, panmodel, minflux=minflux, nsol=1, verbose=False)


        # if empty solution (no reactions to add: models can already grow):
        if first_sol_rids == []: 
            inserted_rids_medium[medium_name] = []
            obj_value_medium[medium_name] = ss_model.optimize().objective_value
            status_medium[medium_name] = ss_model.optimize().status
            

        # if a solution was found:
        elif first_sol_rids != None:
            # add the reactions: 
            for rid in first_sol_rids:
                import_from_universe(ss_model, panmodel, rid, bounds=None, gpr='')
                inserted_rids.append(rid)
            inserted_rids_medium[medium_name] = inserted_rids
            obj_value_medium[medium_name] = ss_model.optimize().objective_value
            status_medium[medium_name] = ss_model.optimize().status


        else:  # if no solution was found: 
            # starting the strenghten_uptakes trick...
            # nested 'with' statement (here + gapfilling) doesn't work, so we create a dictionary to later restore edited bounds:
            exr_ori_ss = strenghten_uptakes(ss_model)
            exr_ori_pan = strenghten_uptakes(panmodel)
            multiplier = 1


            while (first_sol_rids==None and (minflux*multiplier)<= panmodel.slim_optimize()):
                # flux trough the objective could be too low. Starting from the same 'minflux' as before,
                # we try several gapfilling at increasing flux trough the objective. Each iteration raise 1 order of magnitude.
                first_sol_rids = perform_gapfilling(ss_model, panmodel, minflux=minflux*multiplier, nsol=1, verbose=False)
                multiplier = multiplier * 10
            # now restore the medium changes!
            for rid in exr_ori_ss.keys(): ss_model.reactions.get_by_id(rid).lower_bound = exr_ori_ss[rid]
            for rid in exr_ori_pan.keys(): panmodel.reactions.get_by_id(rid).lower_bound = exr_ori_pan[rid]


            # if a solution was found using the strenghten_uptakes trick:
            if first_sol_rids != None:
                # add the needed reactions:
                for rid in first_sol_rids:
                    import_from_universe(ss_model, panmodel, rid, bounds=None, gpr='')
                    inserted_rids.append(rid)
                inserted_rids_medium[medium_name] = inserted_rids
                obj_value_medium[medium_name] = ss_model.optimize().objective_value
                status_medium[medium_name] = ss_model.optimize().status


                # now check if this second gap-filling strategy was enough to reach 'minflux' with the given medium: 
                res = ss_model.optimize()
                obj_value = res.objective_value
                status = res.status
                if status=='optimal' and obj_value < minflux:  # could still be 0


                    # retry the original gap-filling (a kind of polishing): 
                    first_sol_rids = perform_gapfilling(ss_model, panmodel, minflux=minflux, nsol=1, verbose=False)
                    if first_sol_rids != None:  # additional reactions needed to satisfy the thresholds:
                        for rid in first_sol_rids:
                            import_from_universe(ss_model, panmodel, rid, bounds=None, gpr='')
                            inserted_rids.append(rid)
                        inserted_rids_medium[medium_name] = inserted_rids
                        obj_value_medium[medium_name] = ss_model.optimize().objective_value
                        status_medium[medium_name] = ss_model.optimize().status


                    else:  # still no solution despite the strenghten_uptakes trick + the polishing: 
                        gapfilling_failed = medium_name
                        break
            else:  # still no solution despite the strenghten_uptakes trick:
                gapfilling_failed = medium_name
                break
                
    
    # if gapfilling failed on some medium:
    if gapfilling_failed != False: 
        return [{'accession': accession, 'R': '-', 'inserted_rids': '-', 'solver_error': f'failing gapfilling on {gapfilling_failed}', 'obj_value_gf': '-', 'status_gf': '-'}]

    
    # remove disconnected metabolites right before saving the gapfilled model: 
    to_remove = []
    for m in ss_model.metabolites:
        if len(m.reactions) == 0: 
            to_remove.append(m)
    ss_model.remove_metabolites(to_remove)
    
    
    # save strain specific model to disk
    n_R = len(ss_model.reactions)
    cobra.io.save_json_model(ss_model, f'{outdir}/{accession}.json')
    if sbml: cobra.io.write_sbml_model(ss_model, f'{outdir}/{accession}.xml')
    
    
    # compose the new row:
    return [{'accession': accession, 'R': n_R, 'inserted_rids': inserted_rids_medium, 'solver_error': '-', 'obj_value_gf': obj_value_medium, 'status_gf': status_medium}]



def get_gapfilling_matrix(results_df, outdir):
    
    
    gf_matrix = []  # list of dictionaries future dataframe
    for accession in results_df.index: 
        inserted_rids = results_df.loc[accession, 'inserted_rids']
        
        
        # populate the tabular results:
        if type(inserted_rids) == str:
            if inserted_rids == '-':  # model not gapfilled.
                gf_matrix.append({'accession': accession})
        else: 
            row_dict = {}
            for rid in inserted_rids:
                row_dict[rid] = 1
            row_dict['accession'] = accession
            gf_matrix.append(row_dict)
            
    
    # convert to dataframe: 
    gf_matrix = pnd.DataFrame.from_records(gf_matrix)
    gf_matrix = gf_matrix.set_index('accession', drop=True)
    gf_matrix = gf_matrix.fillna(0)  # Replace missing values with 0.
    gf_matrix = gf_matrix.astype(int)  # Force from float to int.
    
    
    # save to file:
    gf_matrix.to_csv(outdir + 'gf_matrix.csv')
            


def strain_filler(logger, outdir, cores, panmodel, media_filepath, minflux, sbml):
    
    
    # log some messages:
    logger.info("Gap-filling strain-specific models...")

   
    # create output dir
    if os.path.exists(outdir + 'strain_models_gf/'):
        # always overwriting if already existing
        shutil.rmtree(outdir + 'strain_models_gf/')  
    os.makedirs(outdir + 'strain_models_gf/', exist_ok=True)
    
    
    # get the list of media on which to gap-fill: 
    media = get_media_definitions(logger, media_filepath)
    if type(media)==int: return 1   # we encountered an error.


    # check if panmodel can grow on the provided media
    if check_panmodel_growth(logger, panmodel, media, minpanflux=0.001) == False:
        return 1
    

    # create items for parallelization: 
    items = []
    for file in glob.glob(outdir + 'strain_models/*.json'):
        items.append(file)
        
        
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession', 'R', 'solver_error']), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_strainfiller),  # will return a new sequences dataframe (to be concat).
            itertools.repeat({'panmodel': panmodel, 'minflux': minflux, 'outdir': outdir + 'strain_models_gf', 'media': media, 'sbml': sbml}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    # join with the previous table, and save:
    results_df = pnd.read_csv(outdir + 'derive_strains.csv', index_col=0)
    results_df = pnd.concat([results_df, all_df_combined], axis=1)
    results_df = results_df.sort_values(by='accession')
    results_df.to_csv(outdir + 'derive_strains.csv')
    
    
    # create the gapfilling matrix starting from 'results_df'
    get_gapfilling_matrix(results_df, outdir)
    
    
    return 0